import os
import numpy as np
import matplotlib.pyplot as plt

from PyQt5.QtWidgets import (
    QCheckBox,
    QFileDialog,
    QMainWindow,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QWidget,
    QHBoxLayout,
    QGroupBox,
    QGridLayout,
    QSpacerItem,
    QMessageBox,
)

from PyQt5.QtGui import QIntValidator
from PyQt5.QtCore import pyqtSignal, QSettings

from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import pandas as pd

from pyboat import core
from pyboat import plotting as pl
from pyboat.ui.util import posfloatV, mkGenericCanvas, selectFilter

FormatFilter = "csv ( *.csv);; MS Excel (*.xlsx);; Text File (*.txt)"


class mkTimeSeriesCanvas(FigureCanvas):

    # dpi != 100 looks wierd?!
    def __init__(self, parent=None, width=4, height=3, dpi=100):

        self.fig1 = Figure(figsize=(width, height), dpi=dpi)

        FigureCanvas.__init__(self, self.fig1)
        self.setParent(parent)

        # print ('Time Series Size', FigureCanvas.sizeHint(self))
        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)


class FourierAnalyzer(QMainWindow):
    def __init__(self, signal, dt, signal_id, position, time_unit, show_T, parent=None):
        super().__init__(parent=parent)

        self.time_unit = time_unit
        self.show_T = show_T

        # --- calculate Fourier spectrum ------------------
        self.fft_freqs, self.fpower = core.compute_fourier(signal, dt)
        # -------------------------------------------------

        self.initUI(position, signal_id)

    def initUI(self, position, signal_id):

        self.setWindowTitle("Fourier spectrum " + signal_id)
        self.setGeometry(510 + position, 80 + position, 520, 600)

        main_frame = QWidget()
        self.fCanvas = mkFourierCanvas()
        self.fCanvas.setParent(main_frame)
        ntb = NavigationToolbar(self.fCanvas, main_frame)

        # plot it
        ax = pl.mk_Fourier_ax(self.fCanvas.fig, self.time_unit, self.show_T)
        pl.Fourier_spec(ax, self.fft_freqs, self.fpower, self.show_T)
        self.fCanvas.fig.subplots_adjust(left=0.15)
        # self.fCanvas.fig.tight_layout()

        main_layout = QGridLayout()
        main_layout.addWidget(self.fCanvas, 0, 0, 9, 1)
        main_layout.addWidget(ntb, 10, 0, 1, 1)

        main_frame.setLayout(main_layout)
        self.setCentralWidget(main_frame)
        self.show()


class mkFourierCanvas(FigureCanvas):
    def __init__(self):
        self.fig, self.ax = plt.subplots(1, 1)

        FigureCanvas.__init__(self, self.fig)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)


class WaveletAnalyzer(QMainWindow):
    def __init__(
        self,
        signal,
        dt,
        Tmin,
        Tmax,
        position,
        signal_id,
        step_num,
        pow_max,
        time_unit,
        parent,
        DEBUG=False,
    ):

        super().__init__(parent=parent)

        self.DEBUG = DEBUG

        self.signal_id = signal_id
        self.signal = signal
        self.pow_max = pow_max
        self.time_unit = time_unit

        self.periods = np.linspace(Tmin, Tmax, step_num)

        # generate time vector
        self.tvec = np.arange(0, len(signal)) * dt

        # no ridge yet
        self.ridge = None
        self.ridge_data = None
        self.power_thresh = None
        self.rsmoothing = None
        self._has_ridge = False  # no plotted ridge

        # no anneal parameters yet
        self.anneal_pars = None

        # =============Compute Spectrum========================================
        self.modulus, self.wlet = core.compute_spectrum(self.signal, dt, self.periods)
        # =====================================================================

        # Wavelet ridge-readout results
        self.ResultWindows = {}
        self.w_offset = 0

        self.initUI(position)

    def initUI(self, position):
        self.setWindowTitle("Wavelet Spectrum - " + str(self.signal_id))
        self.setGeometry(510 + position, 80 + position, 620, 750)

        main_widget = QWidget()
        self.statusBar()

        # Wavelet and signal plot
        self.wCanvas = mkWaveletCanvas()
        main_frame = QWidget()
        self.wCanvas.setParent(main_frame)
        ntb = NavigationToolbar(self.wCanvas, main_frame)  # full toolbar

        # -------------plot the wavelet power spectrum---------------------------

        # creates the ax and attaches it to the widget figure
        axs = pl.mk_signal_modulus_ax(self.time_unit, fig=self.wCanvas.fig)

        pl.plot_signal_modulus(
            axs,
            time_vector=self.tvec,
            signal=self.signal,
            modulus=self.modulus,
            periods=self.periods,
            p_max=self.pow_max,
        )

        self.wCanvas.fig.subplots_adjust(bottom=0.11, right=0.95, left=0.13, top=0.95)
        self.wCanvas.fig.tight_layout()

        # --- Spectrum plotting options ---

        spectrum_opt_box = QGroupBox("Spectrum Plotting Options")
        spectrum_opt_box.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        spectrum_opt_layout = QHBoxLayout()
        spectrum_opt_layout.setSpacing(10)
        spectrum_opt_box.setLayout(spectrum_opt_layout)

        # uppler limit of the colormap <-> imshow(...,vmax = pmax)
        pmax_label = QLabel("Maximal Power:")
        self.pmax_edit = QLineEdit()
        self.pmax_edit.setStatusTip("Sets upper power limit of the spectrum")

        self.pmax_edit.setMaximumWidth(80)
        self.pmax_edit.setValidator(posfloatV)

        # retrieve initial power value, axs[1] is the spectrum
        pmin, pmax = axs[1].images[0].get_clim()
        self.pmax_edit.insert(f"{pmax:.0f}")

        RePlotButton = QPushButton("Update Plot", self)
        RePlotButton.setStatusTip(
            "Rescales the color map of the spectrum with the new max value"
        )
        RePlotButton.clicked.connect(self.update_plot)

        self.cb_coi = QCheckBox("COI", self)
        self.cb_coi.setStatusTip("Draws the cone of influence onto the spectrum")
        self.cb_coi.stateChanged.connect(self.draw_coi)

        # ridge_opt_layout.addWidget(drawRidgeButton,1,3) # not needed anymore?!
        spectrum_opt_layout.addWidget(pmax_label)
        spectrum_opt_layout.addWidget(self.pmax_edit)
        spectrum_opt_layout.addStretch(0)
        spectrum_opt_layout.addWidget(RePlotButton)
        spectrum_opt_layout.addStretch(0)
        spectrum_opt_layout.addWidget(self.cb_coi)

        # --- Time average -> asymptotic Fourier ---

        time_av_box = QGroupBox("Time Averaging")

        estimFourierButton = QPushButton("Estimate Fourier", self)
        estimFourierButton.clicked.connect(self.ini_average_spec)
        estimFourierButton.setStatusTip("Shows time averaged Wavelet power spectrum")

        time_av_layout = QHBoxLayout()
        # time_av_layout.addStretch()
        time_av_layout.addWidget(estimFourierButton)
        # time_av_layout.addStretch()

        time_av_box.setLayout(time_av_layout)

        # --- Ridge detection and options --

        ridge_opt_box = QGroupBox("Ridge Detection")
        ridge_opt_box.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        ridge_opt_layout = QGridLayout()
        ridge_opt_box.setLayout(ridge_opt_layout)

        # Start ridge detection
        maxRidgeButton = QPushButton("Detect Maximum Ridge", self)
        maxRidgeButton.setStatusTip("Finds the time-consecutive power maxima")

        maxRidgeButton.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        maxRidgeButton.clicked.connect(self.do_maxRidge_detection)

        # remove annealing, too slow.. not well implemented
        # annealRidgeButton = QPushButton('Set up ridge\nfrom annealing', self)
        # annealRidgeButton.clicked.connect(self.set_up_anneal)

        power_label = QLabel("Ridge Threshold:")
        power_label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        power_thresh_edit = QLineEdit()
        power_thresh_edit.setStatusTip(
            "Sets the minimal power value required to be considered part of the ridge"
        )
        power_thresh_edit.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        power_thresh_edit.setMinimumSize(60, 0)
        power_thresh_edit.setValidator(posfloatV)

        smooth_label = QLabel("Ridge Smoothing:")
        ridge_smooth_edit = QLineEdit()
        ridge_smooth_edit.setStatusTip(
            "Savitzky-Golay smoothing (k=3) of the ridge in time"
        )
        ridge_smooth_edit.setMinimumSize(60, 0)
        ridge_smooth_edit.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        ridge_smooth_edit.setValidator(QIntValidator(bottom=3, top=len(self.signal)))

        # Plot Results
        plotResultsButton = QPushButton("Plot Ridge Readout", self)
        plotResultsButton.setStatusTip(
            "Shows instantaneous period, phase, power and amplitude along the ridge"
        )
        # plotResultsButton.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        plotResultsButton.clicked.connect(self.ini_plot_readout)

        ridge_opt_layout.addWidget(maxRidgeButton, 0, 0, 1, 1)
        ridge_opt_layout.addWidget(plotResultsButton, 1, 0, 1, 1)
        # ridge_opt_layout.addWidget(annealRidgeButton,1,0)

        ridge_opt_layout.addWidget(power_label, 0, 1)
        ridge_opt_layout.addWidget(power_thresh_edit, 0, 2)

        ridge_opt_layout.addWidget(smooth_label, 1, 1)
        ridge_opt_layout.addWidget(ridge_smooth_edit, 1, 2)

        # for spacing
        rtool_box = QWidget()
        rtool_layout = QHBoxLayout()
        rtool_box.setLayout(rtool_layout)
        rtool_layout.addWidget(ridge_opt_box)
        rtool_layout.addStretch(0)

        # --- put everything together ---

        main_layout = QGridLayout()
        # main_layout.setSpacing(0)
        main_layout.addWidget(self.wCanvas, 0, 0, 4, 5)
        main_layout.addWidget(ntb, 5, 0, 1, 5)
        main_layout.addWidget(spectrum_opt_box, 6, 0, 1, 3)
        main_layout.addWidget(time_av_box, 6, 3, 1, 1)
        main_layout.addItem(QSpacerItem(2, 15), 6, 5)
        main_layout.addWidget(ridge_opt_box, 7, 0, 1, 5)

        # set stretching (resizing) behavior
        main_layout.setRowStretch(0, 1)  # plot should stretch
        main_layout.setRowMinimumHeight(0, 300)  # plot should not get squeezed too much

        main_layout.setRowStretch(5, 0)  # options shouldn't stretch
        main_layout.setRowStretch(6, 0)  # options shouldn't stretch
        main_layout.setRowStretch(7, 0)  # options shouldn't stretch

        # initialize line edits

        power_thresh_edit.textChanged[str].connect(self.qset_power_thresh)
        power_thresh_edit.insert("0.0")  # initialize with 0

        ridge_smooth_edit.textChanged[str].connect(self.qset_ridge_smooth)
        ridge_smooth_edit.insert("0")  # initialize with 0

        main_widget.setLayout(main_layout)
        self.setCentralWidget(main_widget)
        self.show()

    def qset_power_thresh(self, text):

        # catch empty line edit
        if not text:
            return
        text = text.replace(",", ".")

        try:
            power_thresh = float(text)
            self.power_thresh = power_thresh
        # no parsable input
        except ValueError:
            return

        if self.DEBUG:
            print("power thresh set to: ", self.power_thresh)

        # update the plot on the fly
        if self._has_ridge:
            self.draw_ridge()

    def qset_ridge_smooth(self, text):

        """
        rsmooth is the window size for
        the savgol filter
        """

        # catch empty line edit
        if not text:
            return

        text = text.replace(",", ".")
        try:
            rsmooth = float(text)
            rsmooth = int(text)
        # no parsable input
        except ValueError:
            return

        # make an odd window length
        if rsmooth == 0:
            self.rsmoothing = None
        elif rsmooth < 5:
            self.rsmoothing = 5
        elif rsmooth > 5 and rsmooth % 2 == 0:
            self.rsmoothing = rsmooth + 1
        else:
            self.rsmoothing = rsmooth

        # update the plot on the fly
        if self._has_ridge:
            self.draw_ridge()

        if self.DEBUG:
            print("ridge smooth win_len set to: ", self.rsmoothing)

    def do_maxRidge_detection(self):

        ridge_y = core.get_maxRidge_ys(self.modulus)
        self.ridge = ridge_y

        if not np.any(ridge_y):
            msgBox = QMessageBox()
            msgBox.setWindowTitle("Ridge detection error")
            msgBox.setText("No ridge found..check spectrum!")
            msgBox.exec()

            return

        self._has_ridge = True
        self.draw_ridge()  # ridge_data made here

    def draw_ridge(self):

        """ makes also the ridge_data !! """

        if not self._has_ridge:

            msgBox = QMessageBox()
            msgBox.setWindowTitle("No Ridge")
            msgBox.setText("Do a ridge detection first!")
            msgBox.exec()

        ridge_data = core.eval_ridge(
            self.ridge,
            self.wlet,
            self.signal,
            self.periods,
            self.tvec,
            power_thresh=self.power_thresh,
            smoothing_wsize=self.rsmoothing,
        )

        # plot the ridge
        ax_spec = self.wCanvas.fig.axes[1]  # the spectrum

        # already has a plotted ridge
        if ax_spec.lines:
            ax_spec.lines.clear()  # remove old ridge line
            self.cb_coi.setCheckState(0)  # remove COI

        pl.draw_Wavelet_ridge(ax_spec, ridge_data, marker_size=1.5)

        # refresh the canvas
        self.wCanvas.draw()

        self.ridge_data = ridge_data

    def update_plot(self):

        """
        Replots the entire spectrum canvas
        with a new maximal power.
        """

        # remove the old plot
        self.wCanvas.fig.clf()

        # retrieve new pmax value
        text = self.pmax_edit.text()
        text = text.replace(",", ".")
        pmax = float(text)  # pmax_edit has a positive float validator

        if self.DEBUG:
            print(f"new pmax value {pmax}")

        # creates the ax and attaches it to the widget figure
        axs = pl.mk_signal_modulus_ax(self.time_unit, fig=self.wCanvas.fig)

        pl.plot_signal_modulus(
            axs,
            time_vector=self.tvec,
            signal=self.signal,
            modulus=self.modulus,
            periods=self.periods,
            p_max=pmax,
        )

        # redraw COI if checkbox is checked
        self.draw_coi()

        # re-draw ridge
        if self._has_ridge:
            self.draw_ridge()

        # refresh the canvas
        self.wCanvas.draw()
        self.wCanvas.show()

    def set_up_anneal(self):

        """Spawns a new AnnealConfigWindow
        deactivated for the public version..!
        """

        if self.DEBUG:
            print("set_up_anneal called")

        # is bound to parent Wavelet Window
        self.ac = AnnealConfigWindow(self, self.DEBUG)
        self.ac.initUI(self.periods)

    def do_annealRidge_detection(self, anneal_pars):

        """Gets called from the AnnealConfigWindow
        deactivated for the public version..!
        """

        if anneal_pars is None:
            return

        # todo add out-of-bounds parameter check in config window
        ini_per = anneal_pars["ini_per"]
        ini_T = anneal_pars["ini_T"]
        Nsteps = int(anneal_pars["Nsteps"])
        max_jump = int(anneal_pars["max_jump"])
        curve_pen = anneal_pars["curve_pen"]

        # get modulus index of initial straight line ridge
        y0 = np.where(self.periods < ini_per)[0][-1]

        ridge_y, cost = core.find_ridge_anneal(
            self.modulus, y0, ini_T, Nsteps, mx_jump=max_jump, curve_pen=curve_pen
        )

        self.ridge = ridge_y

        # draw the ridge and make ridge_data
        self._has_ridge = True
        self.draw_ridge()

    def draw_coi(self):

        """
        Draws the COI on the spectrum.
        Also redraws the ridge!
        """

        ax_spec = self.wCanvas.fig.axes[1]  # the spectrum axis

        # COI desired?
        if bool(self.cb_coi.checkState()):

            # draw on the spectrum
            pl.draw_COI(ax_spec, self.tvec)

        else:
            ax_spec.lines.clear()  # remove coi, and ridge!
            if self._has_ridge:
                self.draw_ridge()  # re-draw ridge

        # refresh the canvas
        self.wCanvas.draw()

    def ini_plot_readout(self):

        if not self._has_ridge:

            msgBox = QMessageBox()
            msgBox.setWindowTitle("No Ridge")
            msgBox.setText("Do a ridge detection first!")
            msgBox.exec()
            return

        # to keep the line shorter..
        wo = self.w_offset
        self.ResultWindows[wo] = WaveletReadoutWindow(
            self.signal_id,
            self.ridge_data,
            time_unit=self.time_unit,
            draw_coi=self.cb_coi.isChecked(),
            pos_offset=self.w_offset,
            DEBUG=self.DEBUG,
            parent=self,
        )
        self.w_offset += 20

    def ini_average_spec(self):

        self.avWspecWindow = AveragedWaveletWindow(self.modulus, parent=self)


class mkWaveletCanvas(FigureCanvas):
    def __init__(self, parent=None):  # , width=6, height=3, dpi=100):

        # dpi changes fontsize, so bette leave it as is..
        self.fig = Figure(dpi=100)

        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)


class WaveletReadoutWindow(QMainWindow):
    def __init__(
        self,
        signal_id,
        ridge_data,
        time_unit,
        draw_coi,
        parent,
        pos_offset=0,
        DEBUG=False,
    ):
        super().__init__(parent=parent)

        self.signal_id = signal_id

        self.draw_coi = draw_coi
        self.ridge_data = ridge_data
        self.time_unit = time_unit

        # creates self.rCanvas and plots the results
        self.initUI(pos_offset)

        self.DEBUG = DEBUG

    def initUI(self, position):

        self.setWindowTitle("Wavelet Results - " + str(self.signal_id))
        self.setGeometry(700 + position, 260 + position, 750, 500)

        # embed the plotting canvas

        self.rCanvas = mkReadoutCanvas()
        main_frame = QWidget()
        self.rCanvas.setParent(main_frame)
        ntb = NavigationToolbar(self.rCanvas, main_frame)

        # --- plot the wavelet results ---------
        pl.plot_readout(
            self.ridge_data,
            self.time_unit,
            fig=self.rCanvas.fig,
            draw_coi=self.draw_coi,
        )
        self.rCanvas.fig.subplots_adjust(
            wspace=0.3, left=0.1, top=0.98, right=0.95, bottom=0.15
        )

        # messes things up here :/
        # self.rCanvas.fig.tight_layout()

        main_layout = QGridLayout()
        main_layout.addWidget(self.rCanvas, 0, 0, 9, 1)
        main_layout.addWidget(ntb, 10, 0, 1, 1)

        # add the save Button
        SaveButton = QPushButton("Save Results", self)
        SaveButton.clicked.connect(self.save_out)

        button_layout_h = QHBoxLayout()
        button_layout_h.addStretch(1)
        button_layout_h.addWidget(SaveButton)
        button_layout_h.addStretch(1)
        main_layout.addLayout(button_layout_h, 11, 0, 1, 1)

        main_frame.setLayout(main_layout)
        self.setCentralWidget(main_frame)
        self.show()

    def save_out(self):

        dialog = QFileDialog()
        dialog.setDefaultSuffix("csv")
        # retrieve or initialize directory path
        settings = QSettings()
        dir_path = settings.value("dir_name", os.path.curdir)
        data_format = settings.value("data_format", "csv")

        # ----------------------------------------------------------
        base_name = str(self.signal_id).replace(" ", "-")
        default_name = os.path.join(dir_path, base_name + "_ridgeRO.")
        default_name += data_format
        # -----------------------------------------------------------
        file_name, sel_filter = dialog.getSaveFileName(
            self,
            "Save ridge readout as",
            default_name,
            FormatFilter,
            selectFilter[data_format],
        )

        # dialog cancelled
        if not file_name:
            return

        file_ext = file_name.split(".")[-1]

        if self.DEBUG:
            print("selected filter:", sel_filter)
            print("out-path:", file_name)
            print("extracted extension:", file_ext)
            print("ridge data keys:", self.ridge_data.keys())

        if file_ext not in ["txt", "csv", "xlsx"]:

            msgBox = QMessageBox()
            msgBox.setWindowTitle("Unknown File Format")
            msgBox.setText("Please append .txt, .csv or .xlsx to the file name!")
            msgBox.exec()

            return

        # the write out calls
        settings = QSettings()
        float_format = settings.value("float_format", "%.3f")

        if file_ext == "txt":
            self.ridge_data.to_csv(
                file_name, index=False, sep="\t", float_format=float_format
            )

        elif file_ext == "csv":
            self.ridge_data.to_csv(
                file_name, index=False, sep=",", float_format=float_format
            )

        elif file_ext == "xlsx":
            self.ridge_data.to_excel(file_name, index=False, float_format=float_format)

        else:
            if self.DEBUG:
                print("Something went wrong during save out..")
            return
        if self.DEBUG:
            print("Saved!")


class mkReadoutCanvas(FigureCanvas):
    def __init__(self):

        self.fig = Figure(figsize=(8.5, 7), dpi=100)

        FigureCanvas.__init__(self, self.fig)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)


class AveragedWaveletWindow(QMainWindow):
    def __init__(self, modulus, parent):
        super().__init__(parent=parent)

        # --- calculate time averaged power spectrum <-> Fourier estimate ---
        self.avWspec = np.sum(modulus, axis=1) / modulus.shape[1]
        # -------------------------------------------------------------------

        # the Wavelet analysis window spawning *this* Widget
        self.parentWA = parent
        self.DEBUG = parent.DEBUG
        self.initUI()

    def initUI(self):

        self.setWindowTitle(f"Fourier Spectrum Estimate - {self.parentWA.signal_id}")
        self.setGeometry(510, 80, 550, 400)

        main_frame = QWidget()
        pCanvas = mkGenericCanvas()
        pCanvas.setParent(main_frame)
        ntb = NavigationToolbar(pCanvas, main_frame)

        # plot it
        pCanvas.fig.clf()

        pl.averaged_Wspec(
            self.avWspec,
            self.parentWA.periods,
            time_unit=self.parentWA.time_unit,
            fig=pCanvas.fig,
        )

        pCanvas.fig.subplots_adjust(left=0.15, bottom=0.17)

        main_layout = QGridLayout()
        main_layout.addWidget(pCanvas, 0, 0, 9, 1)
        main_layout.addWidget(ntb, 10, 0, 1, 1)

        # add the save Button
        SaveButton = QPushButton("Save Results", self)
        SaveButton.clicked.connect(self.save_out)

        button_layout_h = QHBoxLayout()
        button_layout_h.addStretch(1)
        button_layout_h.addWidget(SaveButton)
        button_layout_h.addStretch(1)
        main_layout.addLayout(button_layout_h, 11, 0, 1, 1)

        main_frame.setLayout(main_layout)
        self.setCentralWidget(main_frame)
        self.show()

    def save_out(self):

        df_out = pd.DataFrame(data=self.avWspec, columns=["power"])
        df_out.index.name = "periods"

        dialog = QFileDialog()
        dialog.setDefaultSuffix("csv")
        # retrieve or initialize directory path
        settings = QSettings()
        dir_path = settings.value("dir_name", os.path.curdir)
        data_format = settings.value("data_format", "csv")

        # --------------------------------------------------------
        base_name = str(self.parentWA.signal_id).replace(" ", "-")
        default_name = os.path.join(dir_path, base_name + "_avWavelet.")
        default_name += data_format
        # -----------------------------------------------------------

        file_name, sel_filter = dialog.getSaveFileName(
            self,
            "Save averaged Wavelet spectrum",
            default_name,
            FormatFilter,
            selectFilter[data_format],
        )

        # dialog cancelled
        if not file_name:
            return

        file_ext = file_name.split(".")[-1]

        if self.DEBUG:
            print("selected filter:", sel_filter)
            print("out-path:", file_name)
            print("extracted extension:", file_ext)

        # the write out calls
        settings = QSettings()
        float_format = settings.value("float_format", "%.3f")

        if file_ext == "txt":
            df_out.to_csv(file_name, index=True, sep="\t", float_format=float_format)

        elif file_ext == "csv":
            df_out.to_csv(file_name, index=True, sep=",", float_format=float_format)

        elif file_ext == "xlsx":
            df_out.to_excel(file_name, index=True, float_format=float_format)

        else:
            if self.DEBUG:
                print("Something went wrong during save out..")
            return
        if self.DEBUG:
            print("Saved!")


# --- Not used in the public version.. ---


class AnnealConfigWindow(QWidget):

    """
    Not used in the public version..
    """

    # the signal for the anneal parameters
    signal = pyqtSignal("PyQt_PyObject")

    def __init__(self, parent, DEBUG):

        super().__init__()
        # get properly initialized in set_up_anneal
        self.parentWaveletWindow = parent
        self.DEBUG = DEBUG

    def initUI(self, periods):
        self.setWindowTitle("Ridge from Simulated Annealing")
        self.setGeometry(310, 330, 350, 200)

        config_grid = QGridLayout()

        ini_per = QLineEdit(
            str(int(np.mean(periods)))
        )  # start at middle of period interval
        ini_T = QLineEdit("10")
        Nsteps = QLineEdit("5000")
        max_jump = QLineEdit("3")
        curve_pen = QLineEdit("0")

        # for easy readout and emission
        self.edits = {
            "ini_per": ini_per,
            "ini_T": ini_T,
            "Nsteps": Nsteps,
            "max_jump": max_jump,
            "curve_pen": curve_pen,
        }

        per_ini_lab = QLabel("Initial period guess")
        T_ini_lab = QLabel("Initial temperature")
        Nsteps_lab = QLabel("Number of iterations")
        max_jump_lab = QLabel("Maximal jumping distance")
        curve_pen_lab = QLabel("Curvature cost")

        per_ini_lab.setWordWrap(True)
        T_ini_lab.setWordWrap(True)
        Nsteps_lab.setWordWrap(True)
        max_jump_lab.setWordWrap(True)
        curve_pen_lab.setWordWrap(True)

        OkButton = QPushButton("Run!", self)
        OkButton.clicked.connect(self.read_emit_parameters)

        # 1st column
        config_grid.addWidget(per_ini_lab, 0, 0, 1, 1)
        config_grid.addWidget(ini_per, 0, 1, 1, 1)
        config_grid.addWidget(T_ini_lab, 1, 0, 1, 1)
        config_grid.addWidget(ini_T, 1, 1, 1, 1)
        config_grid.addWidget(curve_pen_lab, 2, 0, 1, 1)
        config_grid.addWidget(curve_pen, 2, 1, 1, 1)

        # 2nd column
        config_grid.addWidget(Nsteps_lab, 0, 2, 1, 1)
        config_grid.addWidget(Nsteps, 0, 3, 1, 1)
        config_grid.addWidget(max_jump_lab, 1, 2, 1, 1)
        config_grid.addWidget(max_jump, 1, 3, 1, 1)
        config_grid.addWidget(OkButton, 2, 3, 1, 1)

        # set main layout
        self.setLayout(config_grid)
        self.show()

    def read_emit_parameters(self):

        anneal_pars = {}
        for par_key in self.edits:
            edit = self.edits[par_key]
            anneal_pars[par_key] = float(edit.text())

        if self.DEBUG:
            print("Anneal pars:", anneal_pars)
        self.parentWaveletWindow.do_annealRidge_detection(anneal_pars)
        # send to WaveletAnalyzer Window
        # self.signal.emit(anneal_pars)



""" Methods to assess oscillatory readouts of an ensemble of signals """

import numpy as np
import pandas as pd

from pyboat.core import find_COI_crossing, complex_average


def average_power_distribution(ridge_results, signal_ids=None, exclude_coi=False):

    """
    Compute the power distribution of an ensemble
    of (ridge-)analyzed signals.

    Central measure is the signal-length (not ridge length!) averaged
    ridge-power of individual signals.

    Parameters
    ----------

    ridge_results : sequence of DataFrames,
                   holds the ridge data for the individual signals.
                   Check pyboat.core.eval_ridge for details
                   about the DataFrame structure

    signal_ids : sequence, optional
                 labels of the analyzed signals, if not given
                 a numeric sequence of len(ridge_results)
                 will be used as labels

    exclude_coi :  bool, if True only average the ridge outside of the COI,
                  for short noisy trajectories there might be no such points!

    Returns
    -------
    power_series : pandas Series with the signal_ids as index
                   and averaged powers as values

    """

    powers = []
    ids = []

    if signal_ids is None:
        signal_ids = np.arange(len(ridge_results))

    assert len(signal_ids) == len(ridge_results)

    # collect the time-averaged powers
    # rd.Nt is the total signal length!
    for rd, _id in zip(ridge_results, signal_ids):

        if exclude_coi:
            # take only ridge power outside of COI
            i_left, i_right = find_COI_crossing(rd)
            mpower = (rd.power[i_left:i_right]).sum() / rd.Nt
        else:
            mpower = rd.power.sum() / rd.Nt

        # can happen if ridge exclusively inside COI
        if not np.isnan(mpower):
            powers.append(mpower)
            ids.append(_id)

    powers_series = pd.Series(index=ids, data=powers)

    # sort by power, descending
    powers_series.sort_values(ascending=False, inplace=True)

    return powers_series


def get_ensemble_dynamics(ridge_results):

    """
    Aggregate all the ridge-readouts (period, amplitude and phase)
    of a ridge-analyzed signal ensemble and return time-continuous median and
    quartiles (as in a time-continuous box plot).
    In the case of phase return the 1st order parameter
    (length of resultant vector)
    as a phase coherence measure over time.

    Signals/ridge-readouts of unequal length in time are Ok!

    Parameters
    ----------

    ridge_results : sequence (or iterable) of DataFrames
                    holding the ridge data for the individual signals
                    check pyboat.core.eval_ridge for details
                    about the DataFrame structure

    Returns
    --------

    A tuple holding 4 data frames, one summary statistic over time
    for period, amplitude, power and phase each.

    """

    # aggregate the observables, missing values because of unequal length
    # get a NaN entry
    periods = pd.concat([r["periods"] for r in ridge_results], axis=1)
    phases = pd.concat([r["phase"] for r in ridge_results], axis=1)
    amplitudes = pd.concat([r["amplitude"] for r in ridge_results], axis=1)
    powers = pd.concat([r["power"] for r in ridge_results], axis=1)

    # median and the quantiles, NaNs get skipped over
    periods_mq1q3 = pd.DataFrame()
    periods_mq1q3["median"] = periods.median(axis=1, skipna=True)
    periods_mq1q3["Q1"] = periods.quantile(q=0.25, axis=1)
    periods_mq1q3["Q3"] = periods.quantile(q=0.75, axis=1)

    amplitudes_mq1q3 = pd.DataFrame()
    amplitudes_mq1q3["median"] = amplitudes.median(axis=1, skipna=True)
    amplitudes_mq1q3["Q1"] = amplitudes.quantile(q=0.25, axis=1)
    amplitudes_mq1q3["Q3"] = amplitudes.quantile(q=0.75, axis=1)

    powers_mq1q3 = pd.DataFrame()
    powers_mq1q3["median"] = powers.median(axis=1, skipna=True)
    powers_mq1q3["Q1"] = powers.quantile(q=0.25, axis=1)
    powers_mq1q3["Q3"] = powers.quantile(q=0.75, axis=1)

    # 1st order parameter, NaNs of DataFrames get masked for numpy functions!
    R, Psi = complex_average(phases, axis=1)
    phases_R = pd.DataFrame()
    phases_R["R"] = R

    return periods_mq1q3, amplitudes_mq1q3, powers_mq1q3, phases_R



import os
import pandas as pd
import matplotlib.pyplot as plt
import csv

from PyQt5.QtWidgets import (
    QFileDialog,
    QLabel,
    QPushButton,
    QSizePolicy,
    QWidget,
    QGridLayout,
)

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from PyQt5.QtCore import Qt, QAbstractTableModel
from PyQt5.QtGui import QDoubleValidator, QIntValidator

from pyboat.core import interpolate_NaNs

# some Qt Validators, they accept floats with ','!
floatV = QDoubleValidator(bottom=-1e16, top=1e16)
posfloatV = QDoubleValidator(bottom=1e-16, top=1e16)
posintV = QIntValidator(bottom=1, top=9999999999)

# --- the analysis parameter dictionary with defaults ---

default_par_dict = {
    "dt": 1,
    "time_unit": "min",
    "cut_off": None,
    "window_size": None,
    "Tmin": None,
    "Tmax": None,
    "nT": 200,
    "pow_max": None,
    "float_format": "%.3f",
    "graphics_format": "png",
    "data_format": "csv",
}

# map data ouput format to QFileDialog Filter
selectFilter = {
    "csv": "csv ( *.csv)",
    "xlsx": "MS Excel (*.xlsx)",
    "txt": "Text File (*.txt)",
}


class MessageWindow(QWidget):

    """
    A generic window do display a message
    and an Ok buttong to close
    """

    def __init__(self, message, title):
        super().__init__()
        self.message = message
        self.title = title
        self.initUI()

    def initUI(self):
        error = QLabel(self.message)
        self.setGeometry(300, 450, 220, 100)  # x,y, xlen, ylen
        self.setWindowTitle(self.title)
        okButton = QPushButton("OK", self)
        okButton.clicked.connect(self.close)
        main_layout_v = QGridLayout()
        main_layout_v.addWidget(error, 0, 0, 1, 3)
        main_layout_v.addWidget(okButton, 1, 1)
        self.setLayout(main_layout_v)
        self.show()


class mkGenericCanvas(FigureCanvas):
    def __init__(self):
        self.fig, self.ax = plt.subplots(1, 1)

        FigureCanvas.__init__(self, self.fig)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)


def load_data(dir_path="./", debug=False, **kwargs):

    """
    This is the entry point to import the data

    **kwargs: keyword arguments for pandas.read_... functions
    to control things like *header*, *sep* and so on.

    If no 'sep' is present in kwargs, infer column separator from
    default extensions:

    'csv' : ','
    'tsv' : '\t'
    'txt' : '\s+'
    ['xlsx', 'xls'] : use pandas read_excel

    any other extension calls Python's csv.Sniffer in good faith ;)

    """

    err_msg1 = "Non-numeric values encountered in\n\n"
    err_msg2 = "Parsing errors encountered in\n\n"
    err_suffix = "\n\ncheck input..!"

    # returns a list with 1 element, stand alone File Dialog
    file_names = QFileDialog.getOpenFileName(
        parent=None, caption="Open Tabular Data", directory=dir_path
    )

    if debug:
        print(file_names, type(file_names))

    file_name = file_names[0]
    file_ext = file_name.split(".")[-1]

    # check if cancelled -> Null string
    if file_name == "":
        return None, "cancelled", None

    print("Loading", file_ext, file_name)
    new_dir_path = os.path.dirname(file_name)
    # check if it's an excel file:
    if file_ext in ["xls", "xlsx"]:

        # not needed for reading excel sheets
        if "sep" in kwargs:
            del kwargs["sep"]

        try:
            raw_df = pd.read_excel(file_name, **kwargs)
            san_df = sanitize_df(raw_df, debug)
            if san_df is None:
                print("Error loading data..")
                return None, f"{err_msg1}{file_name}{err_suffix}", new_dir_path
            else:
                # attach a name for later reference in the DataViewer
                # strip off extension
                bname = os.path.basename(file_name)
                san_df.name = "".join(bname.split(".")[:-1])
                return san_df, "", os.path.dirname(file_name)  # empty error msg

        except (pd.errors.ParserError, csv.Error):
            return None, f"{err_msg2}{file_name}{err_suffix}", new_dir_path

    # infer table type from extension
    if "sep" not in kwargs:

        if file_ext == "csv":
            delimiter = ","
        elif file_ext == "tsv":
            delimiter = "\t"
        elif file_ext == "txt":
            delimiter = "\s+"
        else:
            # calls Python's inbuilt csv.Sniffer
            delimiter = None
            kwargs["engine"] = "python"

        kwargs["sep"] = delimiter

        if debug:
            print(f"Infered separator from extension: {delimiter}")
            print(f"Loading with kwargs: {kwargs}")
    else:
        if debug:
            print(f"Separator was given as: '{kwargs['sep']}'")

    try:
        raw_df = pd.read_table(file_name, **kwargs)
        if raw_df is None:
            print("Error loading data..")
            return None, f"{err_msg1}{file_name}{err_suffix}", new_dir_path

        san_df = sanitize_df(raw_df, debug)
        if san_df is None:
            print("Error sanitizing data..")
            return None, f"{err_msg1}{file_name}{err_suffix}", new_dir_path

        else:
            # attach a name for later reference in the DataViewer
            bname = os.path.basename(file_name)
            san_df.name = "".join(bname.split(".")[:-1])
            return san_df, "", new_dir_path  # empty error msg

    except pd.errors.ParserError:
        return None, f"{err_msg2}{file_name}{err_suffix}", new_dir_path


def sanitize_df(raw_df, debug=False):

    """
    Makes sure the DataFrame is in a form to
    work as a 'PandasModel', see class below
    """

    # catch wrongly parsed input data, we want all numeric!
    for dtype in raw_df.dtypes:
        if dtype == object:
            return None

    # map all columns to strings for the PandasModel and SignalBox
    str_col = map(str, raw_df.columns)
    raw_df.columns = str_col

    if debug:
        print("raw columns:", raw_df.columns)

    # return data and empty error message
    return raw_df


def interpol_NaNs(df):

    """
    Calls the NaN interpolator for each column
    of the DataFrame
    """

    df = df.apply(interpolate_NaNs, axis=0)

    return df


class PandasModel(QAbstractTableModel):
    """
    Class to populate a table view with a pandas dataframe
    """

    def __init__(self, data, parent=None):
        QAbstractTableModel.__init__(self, parent)
        self._data = data

    def rowCount(self, parent=None):
        return len(self._data.values)

    def columnCount(self, parent=None):
        return self._data.columns.size

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole:
                return str(self._data.values[index.row()][index.column()])
        return None

    def headerData(self, col, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._data.columns[col]
        return None


def set_max_width(qwidget, width):

    size_pol = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
    qwidget.setSizePolicy(size_pol)
    qwidget.setMaximumWidth(width)
    # qwidget.resize( 10,10 )


def retrieve_double_edit(edit):

    text = edit.text()
    text = text.replace(",", ".")
    try:
        value = float(text)
    except ValueError:
        value = None

    return value



""" This module provides all visualizations, both for the ui and the API """

import matplotlib.pyplot as ppl
import numpy as np
from numpy import pi
from scipy.stats import gaussian_kde, iqr

from pyboat.core import get_COI_branches, find_COI_crossing

# --- define colors ---
rgb_2mpl = lambda R, G, B: np.array((R, G, B)) / 255
rgba_2mpl = lambda R, G, B, A: np.array((R, G, B, A)) / 255

SIG_COLOR = "darkslategray"
TREND_COLOR = rgb_2mpl(165, 105, 189)  # orchidy
ENVELOPE_COLOR = "orange"

DETREND_COLOR = "black"
FOURIER_COLOR = "cadetblue"
RIDGE_COLOR = "crimson"

COI_COLOR = "0.6"  # light gray

# average power histogram
HIST_COLOR = "cadetblue"

# the readouts
PERIOD_COLOR = "cornflowerblue"
PHASE_COLOR = "crimson"
AMPLITUDE_COLOR = "k"
POWER_COLOR = "cadetblue"
POWERKDE_COLOR = rgba_2mpl(10, 10, 10, 160)

# the colormap for the wavelet spectra
CMAP = "YlGnBu_r"
# CMAP = 'cividis'
# CMAP = 'magma'

# --- max size of signal to plot also the sample points with explicit marker o's
Nmax = 250

# --- define line widths ---
TREND_LW = 2.0
SIGNAL_LW = 1.5
MARKER_SIZE = 4

# --- standard sizes ---
label_size = 18
tick_label_size = 16

# size of x-axis in inches to
# match dimensions of spectrum and signal plots
x_size = 6.5

# --- default styles as dictionaries ---
# can be used with ppl.plot(..., **STYLE)

SIGNAL_STYLE = {
    "lw": SIGNAL_LW,
    "marker": None,
    "ms": MARKER_SIZE,
    "color": SIG_COLOR,
    "alpha": 0.8,
}
TREND_STYLE = {
    "lw": TREND_LW,
    "marker": None,
    "ms": MARKER_SIZE,
    "color": TREND_COLOR,
    "alpha": 0.8,
}
ENVELOPE_STYLE = {
    "lw": TREND_LW,
    "marker": None,
    "ms": MARKER_SIZE,
    "color": ENVELOPE_COLOR,
    "alpha": 0.8,
}


# signal plot style, show markers only for short signals
def get_marker_lw(signal):

    if len(signal) <= Nmax:
        m = "o"
        lw = 1.2
    else:
        m = ""
        lw = SIGNAL_LW
    return m, lw


# --- Signal and Trend -----------------------------------------------


def mk_signal_ax(time_unit="a.u.", fig=None):

    if fig is None:
        fig = ppl.figure(figsize=(x_size, 3.2))
        fig.subplots_adjust(bottom=0.18)

    ax = fig.subplots()

    ax.set_xlabel("Time (" + time_unit + ")", fontsize=label_size)
    ax.set_ylabel(r"Signal", fontsize=label_size)
    ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
    ax.yaxis.offsetText.set_fontsize(tick_label_size)
    ax.tick_params(axis="both", labelsize=tick_label_size)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    return ax


def draw_signal(ax, time_vector, signal):

    m, lw = get_marker_lw(signal)

    ax.plot(
        time_vector, signal, lw=lw, marker=m, ms=MARKER_SIZE, color=SIG_COLOR, alpha=0.8
    )


def draw_trend(ax, time_vector, trend, label="Trend"):
    ax.plot(time_vector, trend, color=TREND_COLOR, alpha=0.8, lw=TREND_LW, label=label)


def draw_envelope(ax, time_vector, envelope):
    ax.plot(
        time_vector,
        envelope,
        color=ENVELOPE_COLOR,
        alpha=0.8,
        lw=TREND_LW,
        label="Envelope",
    )


def draw_detrended(ax, time_vector, detrended):

    m, lw = get_marker_lw(detrended)

    ax2 = ax.twinx()
    ax2.spines["top"].set_visible(False)

    ax2.plot(
        time_vector,
        detrended,
        "-",
        marker=m,
        ms=MARKER_SIZE,
        color=DETREND_COLOR,
        lw=lw,
        alpha=0.6,
        label="Detrended",
    )

    ax2.set_ylabel("detrended", fontsize=label_size)
    ax2.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
    ax2.yaxis.offsetText.set_fontsize(tick_label_size)

    return ax2


# --- Fourier Spectrum --------------------------


def mk_Fourier_ax(fig, time_unit="a.u.", show_periods=False):

    fig.clf()
    ax = fig.subplots()

    if show_periods:
        ax.set_xlabel("Period (" + time_unit + ")", fontsize=label_size)
        # ax.set_xscale("log")

    else:
        ax.set_xlabel("Frequency (" + time_unit + r"$^{-1}$)", fontsize=label_size)

    ax.set_ylabel("Fourier power", fontsize=label_size)
    ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
    ax.tick_params(axis="both", labelsize=tick_label_size)
    ax.yaxis.offsetText.set_fontsize(tick_label_size)

    return ax


def Fourier_spec(ax, fft_freqs, fft_power, show_periods=False):

    if show_periods:
        # period view, omit the last bin 2/(N*dt)
        # skip 0-frequency
        if len(fft_freqs) < 1000:
            ax.vlines(
                1 / fft_freqs[1:],
                0,
                fft_power[1:],
                lw=1.5,
                alpha=0.8,
                color=FOURIER_COLOR,
            )
        # plot differently for very long signals
        else:
            ax.plot(
                1 / fft_freqs[1:],
                fft_power[1:],
                "-",
                lw=1.5,
                alpha=0.8,
                color=FOURIER_COLOR,
            )
    else:
        if len(fft_freqs) < 1000:
            # frequency view
            ax.vlines(fft_freqs, 0, fft_power, alpha=0.8, color=FOURIER_COLOR, lw=1.5)
        else:
            ax.plot(fft_freqs, fft_power, "-", lw=1.5, alpha=0.8, color=FOURIER_COLOR)


# --- time averaged Wavelet spectrum -> Fourier estimate


def averaged_Wspec(averaged_Wspec, periods, time_unit="a.u", fig=None):

    """
    Plots the time averaged Wavelet spectrum, which is a good Fourier estimate.

    Parameters
    ----------

    averaged_Wspec : sequence, holding the average power for each period

    periods :   sequence, the periods of the original Wavelet transform

    time_unit : str, the time unit label

    fig : matplotlib figure instance, a new figure is created per default

    """

    if fig is None:
        fig = ppl.figure(figsize=(5, 3.2))

    ax = fig.subplots()
    ax.set_ylabel("Power (wnp)", fontsize=label_size)
    ax.set_xlabel(f"Period ({time_unit})", fontsize=label_size)

    ax.plot(periods, averaged_Wspec, lw=SIGNAL_LW, color=FOURIER_COLOR)
    ax.fill_between(periods, 0, averaged_Wspec, color=FOURIER_COLOR, alpha=0.3)

    return ax


# --- Wavelet spectrum  ------


def mk_signal_modulus_ax(time_unit="a.u.", height_ratios=[1, 2.5], fig=None):

    if fig is None:
        fig = ppl.figure(figsize=(x_size, 6.5))
        fig.subplots_adjust(bottom=0.07, top=0.97)
    # 1st axis is for signal, 2nd axis is the spectrum
    axs = fig.subplots(2, 1, gridspec_kw={"height_ratios": height_ratios}, sharex=True)

    sig_ax = axs[0]
    mod_ax = axs[1]

    sig_ax.set_ylabel("Signal (a.u.)", fontsize=label_size)

    mod_ax.set_xlabel("Time (" + time_unit + ")", fontsize=label_size)
    mod_ax.set_ylabel("Period (" + time_unit + ")", fontsize=label_size)
    mod_ax.tick_params(axis="both", labelsize=tick_label_size)

    return axs


def plot_signal_modulus(axs, time_vector, signal, modulus, periods, p_max=None):

    """
    Plot the signal and the wavelet power spectrum.
    axs[0] is signal axis, axs[1] spectrum axis
    """

    sig_ax = axs[0]
    mod_ax = axs[1]

    # Plot Signal above spectrum

    m, lw = get_marker_lw(signal)

    sig_ax.plot(
        time_vector, signal, color="black", lw=lw, alpha=0.65, marker=m, ms=MARKER_SIZE
    )

    # sig_ax.plot(time_vector, signal, ".", color="black", ms=2.0, alpha=0.5)
    sig_ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
    sig_ax.tick_params(axis="both", labelsize=tick_label_size)
    sig_ax.yaxis.offsetText.set_fontsize(tick_label_size)

    # Plot Wavelet Power Spectrum

    extent = (time_vector[0], time_vector[-1], periods[0], periods[-1])
    im = mod_ax.imshow(
        modulus[::-1], cmap=CMAP, vmin=0, vmax=p_max, extent=extent, aspect="auto"
    )

    mod_ax.set_ylim((periods[0], periods[-1]))
    mod_ax.set_xlim((time_vector[0], time_vector[-1]))
    mod_ax.grid(axis="y", color="0.6", lw=1.0, alpha=0.5)  # vertical grid lines

    if p_max is None:
        cb_ticks = [0, int(np.floor(modulus.max()))]
    else:
        cb_ticks = [0, p_max]

    cb = ppl.colorbar(
        im, ax=mod_ax, orientation="horizontal", fraction=0.08, shrink=0.6, pad=0.22
    )
    cb.set_ticks(cb_ticks)
    cb.ax.set_xticklabels(cb_ticks, fontsize=tick_label_size)
    # cb.set_label('$|\mathcal{W}_{\Psi}(t,T)|^2$',rotation = '0',labelpad = 5,fontsize = 15)
    cb.set_label("Wavelet Power", rotation="0", labelpad=-12, fontsize=0.9 * label_size)


def draw_Wavelet_ridge(ax, ridge_data, marker_size=1.5):

    """
    *ridge_data* comes from core.eval_ridge !
    """

    ax.plot(
        ridge_data.time,
        ridge_data["periods"],
        "o",
        color=RIDGE_COLOR,
        alpha=0.6,
        ms=marker_size,
    )


def draw_COI(ax, time_vector):

    """
    Draw Cone of influence on spectrum, period version

    """

    coi, coi_t = get_COI_branches(time_vector)

    ax.plot(coi_t, coi, "-.", alpha=0.7, color=COI_COLOR)


# --- Wavelet readout ----------------------------


def plot_readout(ridge_data, time_unit="a.u.", draw_coi=False, fig=None):

    """
    ridge_data from core.eval_ridge(...)
    creates four axes: period, phase, amplitude and power
    """

    # restore the original complete time vector
    tvec = np.arange(ridge_data.Nt) * ridge_data.dt

    i_left, i_right = find_COI_crossing(ridge_data)

    if fig is None:
        fig = ppl.figure(figsize=(7, 4.8))

    periods = ridge_data["periods"]
    phases = ridge_data["phase"]
    powers = ridge_data["power"]
    amplitudes = ridge_data["amplitude"]
    ridge_t = ridge_data["time"]  # indexable

    # check for discontinuous ridge
    if np.all(np.diff(ridge_t, n=2) < 1e-12):
        # draw continuous lines
        lstyle = "-"
        mstyle = ""
    else:
        lstyle = ""
        mstyle = "o"

    fig.subplots_adjust(wspace=0.3, left=0.11, top=0.98, right=0.97, bottom=0.14)

    axs = fig.subplots(2, 2, sharex=True)

    for ax in axs.flatten():
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)

    ax1 = axs[0, 0]
    ax2 = axs[0, 1]
    ax3 = axs[1, 0]
    ax4 = axs[1, 1]

    # periods
    ax1.plot(
        ridge_t.loc[i_left:i_right],
        periods.loc[i_left:i_right],
        color=PERIOD_COLOR,
        alpha=0.8,
        lw=2.5,
        ls=lstyle,
        marker=mstyle,
        ms=1.5,
    )
    # inside COI
    ax1.plot(
        ridge_t.loc[:i_left],
        periods.loc[:i_left],
        "--",
        color=PERIOD_COLOR,
        alpha=0.8,
        ms=2.5,
    )
    ax1.plot(
        ridge_t.loc[i_right:],
        periods.loc[i_right:],
        "--",
        color=PERIOD_COLOR,
        alpha=0.8,
        ms=2.5,
    )

    ax1.set_ylabel(f"Period ({time_unit})", fontsize=label_size)
    ax1.grid(True, axis="y")
    yl = ax1.get_ylim()
    ax1.set_ylim((max([0, 0.75 * yl[0]]), 1.25 * yl[1]))

    # only now draw the COI?
    if draw_coi:
        draw_COI(ax1, tvec)

    ax1.tick_params(axis="both", labelsize=tick_label_size)

    # ax1.set_ylim( (120,160) )

    # phases
    ax2.plot(
        ridge_t.loc[i_left:i_right],
        phases.loc[i_left:i_right],
        c=PHASE_COLOR,
        alpha=0.8,
        ls=lstyle,
        marker=mstyle,
        ms=1.5,
    )

    # inside COI
    ax2.plot(
        ridge_t.loc[:i_left],
        phases.loc[:i_left],
        "-.",
        color=PHASE_COLOR,
        alpha=0.5,
        ms=2.5,
    )
    ax2.plot(
        ridge_t.loc[i_right:],
        phases.loc[i_right:],
        "-.",
        color=PHASE_COLOR,
        alpha=0.5,
        ms=2.5,
    )

    ax2.set_ylabel("Phase (rad)", fontsize=label_size, labelpad=0.5)
    ax2.set_yticks((0, pi, 2 * pi))
    ax2.set_yticklabels(("$0$", "$\pi$", "$2\pi$"))
    ax2.tick_params(axis="both", labelsize=tick_label_size)

    # amplitudes
    ax3.plot(
        ridge_t.loc[i_left:i_right],
        amplitudes.loc[i_left:i_right],
        c=AMPLITUDE_COLOR,
        lw=2.5,
        alpha=0.9,
        ls=lstyle,
        marker=mstyle,
        ms=1.5,
    )

    # inside COI
    ax3.plot(
        ridge_t.loc[:i_left],
        amplitudes.loc[:i_left],
        "--",
        color=AMPLITUDE_COLOR,
        alpha=0.6,
        ms=2.5,
    )
    ax3.plot(
        ridge_t.loc[i_right:],
        amplitudes.loc[i_right:],
        "--",
        color=AMPLITUDE_COLOR,
        alpha=0.6,
        ms=2.5,
    )

    ax3.set_ylim((0, 1.1 * amplitudes.max()))
    ax3.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
    ax3.yaxis.offsetText.set_fontsize(tick_label_size)
    ax3.set_ylabel("Amplitude (a.u.)", fontsize=label_size)
    ax3.set_xlabel("Time (" + time_unit + ")", fontsize=label_size)
    ax3.tick_params(axis="both", labelsize=tick_label_size)

    # powers
    ax4.plot(
        ridge_t.loc[i_left:i_right],
        powers.loc[i_left:i_right],
        lw=2.5,
        alpha=0.8,
        color=POWER_COLOR,
        ls=lstyle,
        marker=mstyle,
        ms=1.5,
    )

    # inside COI
    ax4.plot(
        ridge_t.loc[:i_left],
        powers.loc[:i_left],
        "--",
        color=POWER_COLOR,
        alpha=0.6,
        ms=2.5,
    )
    ax4.plot(
        ridge_t.loc[i_right:],
        powers.loc[i_right:],
        "--",
        color=POWER_COLOR,
        alpha=0.6,
        ms=2.5,
    )

    ax4.set_ylim((0, 1.1 * powers.max()))
    ax4.set_ylabel("Power (wnp)", fontsize=label_size)
    ax4.set_xlabel("Time (" + time_unit + ")", fontsize=label_size)
    ax4.tick_params(axis="both", labelsize=tick_label_size)

    fig.tight_layout()

    return axs


# -------- Ensemble Measures Plots -------------------------------------


def Freedman_Diaconis_rule(samples):

    """
    Get optimal number of bins from samples,
    a bit too small in most cases for the power distribution it seems..
    """

    h = 2 * iqr(samples) / pow(samples.size, 1 / 3)
    Nbins = int((samples.max() - samples.min()) / h)
    return Nbins


def Rice_rule(samples):

    """
    Get optimal number of bins from samples, looks about
    right for 'typical' bi-modal power distributions.
    """

    Nbins = int(2 * pow(len(samples), 1 / 3))
    return Nbins


def power_distribution(powers, kde=True, fig=None):

    """
    Histogram (bin-counts) of Wavelet powers, intended
    for the time-averaged powers as computed
    by ensemble_measures.average_power_distribution(...).

    Parameters
    ----------

    powers : a sequence of floats
    kde  : bool, Gaussian kde
    fig :    matplotlib figure instance
    """

    powers = np.array(powers)

    if powers.ndim != 1:
        raise ValueError("Powers must be a sequence!")

    if fig is None:
        fig = ppl.figure(figsize=(5, 3.2))

    ax = fig.subplots()
    ax.set_ylabel("Counts", fontsize=label_size)
    ax.set_xlabel("Average Signal Power", fontsize=label_size)
    ax.tick_params(axis="both", labelsize=tick_label_size)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    # use automated number of bins
    Nbins = Rice_rule(powers)

    # the raw histogram
    counts, bins, _ = ax.hist(
        powers, bins=Nbins, rwidth=0.9, color=HIST_COLOR, alpha=0.7, density=False
    )

    delta_bin = bins[1] - bins[0]  # evenly spaced

    if kde:

        # get the KDE support
        support = np.linspace(0.2 * powers.min(), 1.2 * powers.max(), 200)
        # standard KDE, bandwith from Silverman
        dens = gaussian_kde(powers, bw_method="silverman")

        ax2 = ax.twinx()
        ax2.spines["top"].set_visible(False)
        ax2.spines["right"].set_visible(False)

        # normalize relative to count axis
        ax2.set_ylim((0, ax.get_ylim()[1] / (len(powers) * delta_bin)))
        ax2.set_yticks(())
        ax2.plot(support, dens(support), color=POWERKDE_COLOR, lw=2.0, label="KDE")

        # not needed?!
        # ax2.legend(fontsize = tick_label_size)

    fig.tight_layout()
    # fig.subplots_adjust(bottom = 0.22, left = 0.15)


def ensemble_dynamics(
    periods, amplitudes, powers, phases, dt=1, time_unit="a.u", fig=None
):

    """
    Taking the summary statistics from

    core.get_ensemble_dynamics()

    plots the median and quartiles for the periods
    and amplitudes over time, as well as the
    phase coherence measure R.

    Parameters
    ----------

    periods : DataFrame containing the 'median' and the
              two quartiles 'Q1' and 'Q3' over time
    amplitudes : DataFrame containing the 'median' and the
                 two quartiles 'Q1' and 'Q3' over time

    powers : DataFrame containing the 'median' and the
                 two quartiles 'Q1' and 'Q3' over time

    phases : DataFrame containing 1st order parameter 'R'

    dt : float, the sampling interval to get a proper time axis

    time_unit : str, the time unit label

    """

    # all DataFrames come from the same ensemble, so their shape should match
    tvec = np.arange(periods.shape[0]) * dt

    if fig is None:
        fig = ppl.figure(figsize=(7, 4.8))

    # create the 2 axes
    axs = fig.subplots(2, 2, sharex=True)

    for ax in axs.flatten():
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.grid(axis="y")

    ax1 = axs[0, 0]
    ax2 = axs[0, 1]
    ax3 = axs[1, 0]
    ax4 = axs[1, 1]

    # the periods
    ax1.plot(tvec, periods["median"], c=PERIOD_COLOR, lw=2, alpha=0.7)
    ax1.fill_between(tvec, periods["Q1"], periods["Q3"], color=PERIOD_COLOR, alpha=0.3)
    ax1.set_ylabel(f"Period ({time_unit})", fontsize=label_size)
    ax1.tick_params(axis="both", labelsize=tick_label_size)

    # phase coherence
    ax2.plot(tvec, phases["R"], c=PHASE_COLOR, lw=3, alpha=0.8)

    ax2.set_ylim((0, 1.1))
    ax2.set_ylabel("Phase Coherence", fontsize=label_size)
    ax2.tick_params(axis="both", labelsize=tick_label_size)

    # amplitudes
    ax3.plot(tvec, amplitudes["median"], c=AMPLITUDE_COLOR, lw=2, alpha=0.6)
    ax3.fill_between(
        tvec, amplitudes["Q1"], amplitudes["Q3"], color=AMPLITUDE_COLOR, alpha=0.2
    )
    ax3.set_ylabel("Amplitude (a.u.)", fontsize=label_size)
    ax3.tick_params(axis="both", labelsize=tick_label_size)
    ax3.set_xlabel(f"Time ({time_unit})", fontsize=label_size)

    # powers
    ax4.plot(tvec, powers["median"], c=POWER_COLOR, lw=2, alpha=0.6)
    ax4.fill_between(tvec, powers["Q1"], powers["Q3"], color=POWER_COLOR, alpha=0.2)
    ax4.set_ylabel("Power (a.u.)", fontsize=label_size)
    ax4.tick_params(axis="both", labelsize=tick_label_size)
    ax4.set_xlabel(f"Time ({time_unit})", fontsize=label_size)

    fig.subplots_adjust(wspace=0.3, left=0.11, top=0.98, right=0.97, bottom=0.14)

    # fig.subplots_adjust(bottom = 0.1, left = 0.2, top = 0.95, hspace = 0.1)
    return axs


def Fourier_distribution(
    df_fouriers, time_unit="a.u.", label=None, color=FOURIER_COLOR, fig=None
):

    """
    Plots the median and quartiles of a distribution of
    Fourier power spectra. Typical use case are the
    time averaged Wavelet spectra of a population.

    Parameters
    ----------

    df_fouriers : DataFrame, columns hold the individual Fourier power spectra,
                            index holds the periods

    time_unit : str, the time unit label

    fig : matplotlib figure instance

    """

    periods = df_fouriers.index

    med = df_fouriers.median(axis=1)
    q1 = df_fouriers.quantile(q=0.25, axis=1)
    q3 = df_fouriers.quantile(q=0.75, axis=1)

    if fig is None:
        fig = ppl.figure(figsize=(5.6, 4.5))

    ax = fig.subplots()
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.grid(axis="y")

    ax.tick_params(axis="both", labelsize=tick_label_size)
    ax.set_xlabel(f"Period ({time_unit})", fontsize=label_size)
    ax.set_ylabel("Power (wnp)", fontsize=label_size)

    ax.plot(periods, med, lw=2.5, color=color, label=label)
    ax.fill_between(periods, q1, q3, color=color, alpha=0.3)

    if label:
        ax.legend()

    fig.tight_layout()
    return ax



""" Some simple functions to create synthetic signals mixed with AR(1) noise"""

import numpy as np
from numpy import pi


def ar1_sim(alpha, Nt, sigma=1):

    """
    Simulates an AR(1) stochastic process realization.

    Parameters
    ----------

    alpha : float, The 1st regressive parameter, set to 0 for white noise.
            Must be smalller than 1, otherwise the process diverges!

    Nt : int, Number of samples to be created

    sigma : float, the standard deviation of the gaussian noise
    """

    Nt = int(Nt)
    sol = np.zeros(Nt)

    # pick the 1st value at random
    x0 = np.random.randn()

    sol[0] = x0

    for i in range(1, Nt):
        sol[i] = alpha * sol[i - 1] + sigma * np.random.randn()

    return sol


def create_chirp(T1, T2, Nt):

    """
    Creates a clean chirp signal,
    sweeping linearly through the
    frequencies:

    f_1 = 1/T1, and
    f_2 = 1/T2

    Parameters
    ----------
    T1: positive float, the period at t = 0, unit is sampling interval
    T2: positive float, the period at t = Nt, unit is sampling interval
    Nt: integer, number of sample points
    """

    omega_1, omega_2 = 2 * pi / T1, 2 * pi / T2

    tvec = np.arange(Nt)

    phases = 0.5 / Nt * (omega_2 - omega_1) * tvec ** 2 + omega_1 * tvec

    return np.cos(phases)


def create_noisy_chirp(T1, T2, Nt, eps, alpha=0):

    """
    Creates a noisy chirp signal,
    sweeping linearly through the
    frequencies:

    f_1 = 1/T1, and
    f_2 = 1/T2

    and adds an AR1 realization
    as background noise.

    Paramters
    ---------
    T1: float, starting period, unit is sampling interval
    T1: float, final period, unit is sampling interval
    Nt: int, number of samples
    eps: float, noise intensity
    alpha: float, AR1 parameter 0 < alpha < 1
    """

    signal = create_chirp(T1, T2, Nt)
    noise = ar1_sim(alpha=alpha, Nt=Nt)

    return signal + eps * noise


def create_exp_envelope(tau, Nt):

    """
    Just returns an exponential
    amplitude envelope with
    value 1/e after *tau* time elapsed.

    Parameters
    ----------
    tau: float, in sampling interval units
    nt: integer, number of sample points
    """

    tvec = np.arange(Nt)
    env = np.exp(-1 / tau * tvec)

    return env


def assemble_signal(list_of_components, weights):

    """
    Linearly combines all signal components
    with the given weights.

    list_of_components: list of sequences of same length,
                        the signal components, e.g. chirps, noise, ...
    weights: sequence,
             the weights (amplitude, noise strengths..) of
             each component
    """

    if not len(weights) == len(list_of_components):
        raise ValueError("Need as much weights as signal components!")

    cpts = np.array(list_of_components).T  # time x component
    cpts = weights * cpts

    signal = np.sum(cpts, axis=1)

    return signal


def create_example_trajectory(Nt=500):

    """
    Example signal consisting of a chirp
    with exponential envelope and
    an AR1 background noise
    """

    s1 = create_chirp(T1=20, T2=30, Nt=Nt)
    noise = ar1_sim(alpha=0.3, Nt=Nt)
    env = create_exp_envelope(tau=Nt * 0.7, Nt=Nt)

    signal = assemble_signal([env * s1, noise], [1, 0.5])

    return signal



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from os.path import expanduser

from PyQt5.QtWidgets import (
    QCheckBox,
    QMessageBox,
    QFileDialog,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QWidget,
    QHBoxLayout,
    QGroupBox,
    QGridLayout,
    QProgressBar,
    QSpacerItem,
    QFrame,
    QMainWindow,
)
from PyQt5.QtGui import QIntValidator
from PyQt5.QtCore import QSettings
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar

from pyboat.ui.util import posfloatV, mkGenericCanvas

import pyboat
from pyboat import plotting as pl
from pyboat import ensemble_measures as em


# --- resolution of the output plots ---
DPI = 250


class BatchProcessWindow(QMainWindow):

    """
    The parent is a DataViewer instance holding the
    data as a DataFrame, and other global properties:

    parent.df
    parent.dt
    parent.time_unit

    """

    def __init__(self, parent, DEBUG):

        super().__init__(parent=parent)

        # the DataViewer spawning *this* Widget
        self.parentDV = parent
        self.debug = DEBUG

    def initUI(self, wlet_pars):

        """
        Gets called from the parent DataViewer
        """

        self.setWindowTitle("Batch Processing")
        self.setGeometry(310, 330, 600, 200)

        # from the DataViewer
        self.wlet_pars = wlet_pars

        # for the status bar
        main_widget = QWidget()
        self.statusBar()

        main_layout = QGridLayout()

        # -- Ridge Analysis Options --

        ridge_options = QGroupBox("Ridge Detection")

        thresh_label = QLabel("Ridge Threshold:")
        thresh_edit = QLineEdit()
        thresh_edit.setValidator(posfloatV)
        thresh_edit.insert("0")
        thresh_edit.setMaximumWidth(60)
        thresh_edit.setStatusTip(
            "Ridge points below that power value will be filtered out "
        )
        self.thresh_edit = thresh_edit

        smooth_label = QLabel("Ridge Smoothing:")
        smooth_edit = QLineEdit()
        smooth_edit.setMaximumWidth(60)
        smooth_edit.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        smooth_edit.setValidator(QIntValidator(bottom=3, top=99999999))
        smooth_edit.setStatusTip(
            """Savitkzy-Golay window size for smoothing the ridge,
            leave blank for no smoothing"""
        )
        self.smooth_edit = smooth_edit

        ridge_options_layout = QGridLayout()
        ridge_options_layout.addWidget(thresh_label, 0, 0)
        ridge_options_layout.addWidget(thresh_edit, 0, 1)
        ridge_options_layout.addWidget(smooth_label, 1, 0)
        ridge_options_layout.addWidget(smooth_edit, 1, 1)
        ridge_options.setLayout(ridge_options_layout)

        # -- Plotting Options --

        plotting_options = QGroupBox("Summary Statistics")
        self.cb_power_dis = QCheckBox("Ridge Power Distribution")
        self.cb_power_dis.setStatusTip(
            "Show time-averaged distribution of ridge powers"
        )
        self.cb_plot_ens_dynamics = QCheckBox("Ensemble Dynamics")
        self.cb_plot_ens_dynamics.setStatusTip(
            "Show period, amplitude and phase distribution over time"
        )
        self.cb_plot_Fourier_dis = QCheckBox("Fourier Spectra Distribution")
        self.cb_plot_Fourier_dis.setStatusTip(
            "Ensemble power distribution of the time averaged Wavelet spectra"
        )

        lo = QGridLayout()
        lo.addWidget(self.cb_plot_ens_dynamics, 0, 0)
        lo.addWidget(self.cb_plot_Fourier_dis, 1, 0)
        lo.addWidget(self.cb_power_dis, 2, 0)
        plotting_options.setLayout(lo)

        # -- Save Out Results --

        export_options = QGroupBox("Export Results")
        export_options.setStatusTip("Creates various figures and csv's")
        export_options.setCheckable(True)
        export_options.setChecked(False)

        self.cb_filtered_sigs = QCheckBox("Filtered Signals")
        self.cb_filtered_sigs.setStatusTip(
            "Saves detrended and amplitude normalized signals to disc as csv's"
        )

        self.cb_specs = QCheckBox("Wavelet Spectra")
        self.cb_specs.setStatusTip("Saves the individual wavelet spectra as images")

        self.cb_specs_noridge = QCheckBox("Wavelet Spectra w/o ridges")
        self.cb_specs_noridge.setStatusTip(
            "Saves the individual wavelet spectra without the ridges as images"
        )

        self.cb_readout = QCheckBox("Ridge Readouts")
        self.cb_readout.setStatusTip(
            "Saves one analysis result per signal to disc as csv"
        )

        self.cb_readout_plots = QCheckBox("Ridge Readout Plots")
        self.cb_readout_plots.setStatusTip("Saves the individual readout plots to disc")
        self.cb_sorted_powers = QCheckBox("Sorted Average Powers")
        self.cb_sorted_powers.setStatusTip(
            "Saves the time-averaged ridge powers in descending order"
        )
        self.cb_save_ensemble_dynamics = QCheckBox("Ensemble Dynamics")
        self.cb_save_ensemble_dynamics.setStatusTip(
            "Separately saves period, amplitude, power and phase summary statistics to a csv file"
        )

        self.cb_save_Fourier_dis = QCheckBox("Fourier Distribution")
        self.cb_save_Fourier_dis.setStatusTip(
            "Saves median and quartiles of the ensemble Fourier power spectral distribution"
        )

        # defaults to HOME or former working dir
        # retrieve or initialize directory path
        settings = QSettings()
        dir_path = settings.value("dir_name", expanduser("~"))
        self.OutPath_edit = QLineEdit(dir_path)

        PathButton = QPushButton("Select Path..")
        PathButton.setMaximumWidth(100)
        PathButton.clicked.connect(self.select_export_dir)

        line1 = QFrame()
        line1.setFrameShape(QFrame.HLine)
        line1.setFrameShadow(QFrame.Sunken)

        line2 = QFrame()
        line2.setFrameShape(QFrame.HLine)
        line2.setFrameShadow(QFrame.Sunken)

        lo = QGridLayout()
        lo.setSpacing(0)

        lo.addWidget(self.cb_filtered_sigs, 0, 0)

        lo.addWidget(self.cb_specs, 1, 0)
        lo.addWidget(self.cb_specs_noridge, 2, 0)

        lo.addWidget(self.cb_readout, 3, 0)
        lo.addWidget(self.cb_readout_plots, 4, 0)
        # lo.addWidget(line1, 3,0)
        lo.addWidget(self.cb_sorted_powers, 5, 0)
        lo.addWidget(self.cb_save_ensemble_dynamics, 6, 0)
        lo.addWidget(self.cb_save_Fourier_dis, 7, 0)
        # lo.addWidget(line2, 6,0)
        lo.addWidget(PathButton, 8, 0)
        lo.addWidget(self.OutPath_edit, 9, 0)
        export_options.setLayout(lo)
        self.export_options = export_options

        # -- Progress and Run --
        Nsignals = self.parentDV.df.shape[1]

        RunButton = QPushButton(f"Run for {Nsignals} Signals!", self)
        RunButton.setStyleSheet("background-color: orange")
        RunButton.clicked.connect(self.run_batch)
        # RunButton.setMaximumWidth(60)

        # the progress bar
        self.progress = QProgressBar(self)
        self.progress.setRange(0, Nsignals - 1)
        # self.progress.setGeometry(0,0, 300, 20)
        self.progress.setMinimumWidth(200)

        # nsig_label = QLabel(f'{Nsignals} Signals')

        process_box = QGroupBox("Processing")
        process_box.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        lo = QHBoxLayout()
        lo.addWidget(self.progress)
        lo.addItem(QSpacerItem(30, 2))
        # lo.addStretch(0)
        lo.addWidget(RunButton)
        lo.addStretch(0)
        process_box.setLayout(lo)

        # -- main layout --

        main_layout.addWidget(plotting_options, 0, 0, 1, 1)
        main_layout.addWidget(ridge_options, 1, 0, 1, 1)
        main_layout.addWidget(export_options, 0, 1, 2, 1)
        main_layout.addWidget(process_box, 2, 0, 1, 2)

        # set main layout
        main_widget.setLayout(main_layout)
        self.setCentralWidget(main_widget)

        self.show()

    def run_batch(self):

        """
        Retrieve all batch settings and loop over the signals
        present in the parentDV

        """

        dataset_name = self.parentDV.df.name

        if self.export_options.isChecked():
            OutPath = self.get_OutPath()
            if OutPath is None:
                return

        # TODO: parallelize
        ridge_results, df_fouriers = self.do_the_loop()

        # check for empty ridge_results
        if not ridge_results:

            msgBox = QMessageBox()
            msgBox.setWindowTitle("No Results")
            msgBox.setText("All ridges below threshold.. no results!")
            msgBox.exec()

            return

        settings = QSettings()
        float_format = settings.value("float_format", "%.3f")

        # --- compute the time-averaged powers ---

        if self.cb_power_dis.isChecked() or self.cb_sorted_powers.isChecked():

            powers_series = em.average_power_distribution(
                ridge_results.values(), ridge_results.keys(), exclude_coi=True
            )

        if self.cb_power_dis.isChecked():
            # plot the distribution
            self.pdw = PowerDistributionWindow(powers_series, dataset_name=dataset_name)

        # save out the sorted average powers
        if self.export_options.isChecked() and self.cb_sorted_powers.isChecked():
            fname = os.path.join(OutPath, f"{dataset_name}_ridge-powers.csv")
            powers_series.to_csv(
                fname, sep=",", float_format=float_format, index=True, header=False
            )

        # --- compute summary statistics over time ---

        if (
            self.cb_plot_ens_dynamics.isChecked()
            or self.cb_save_ensemble_dynamics.isChecked()
        ):
            # res is a tuple of  DataFrames, one each for
            # periods, amplitude, power and phase
            res = em.get_ensemble_dynamics(ridge_results.values())

        if self.cb_plot_ens_dynamics.isChecked():
            self.edw = EnsembleDynamicsWindow(
                res,
                dt=self.parentDV.dt,
                time_unit=self.parentDV.time_unit,
                dataset_name=dataset_name,
            )

        if (
            self.export_options.isChecked()
            and self.cb_save_ensemble_dynamics.isChecked()
        ):
            # create time axis, all DataFrames have same number of rows
            tvec = np.arange(res[0].shape[0]) * self.parentDV.dt
            for obs, df in zip(["periods", "amplitudes", "powers", "phasesR"], res):
                fname = os.path.join(OutPath, f"{dataset_name}_{obs}.csv")
                df.index = tvec
                df.index.name = "time"
                df.to_csv(fname, sep=",", float_format=float_format)

        # --- Fourier Distribution Outputs ---

        if self.cb_plot_Fourier_dis.isChecked():

            self.fdw = FourierDistributionWindow(
                df_fouriers, self.parentDV.time_unit, dataset_name
            )

        if self.export_options.isChecked() and self.cb_save_Fourier_dis.isChecked():

            fname = os.path.join(OutPath, f"{dataset_name}_fourier-distribution.csv")

            # save out median and quartiles of Fourier powers
            df_fdis = pd.DataFrame(index=df_fouriers.index)
            df_fdis["Median"] = df_fouriers.median(axis=1)
            df_fdis["Mean"] = df_fouriers.mean(axis=1)
            df_fdis["Q1"] = df_fouriers.quantile(q=0.25, axis=1)
            df_fdis["Q3"] = df_fouriers.quantile(q=0.75, axis=1)

            df_fdis.to_csv(fname, sep=",", float_format=float_format)

        if self.debug:
            print(list(ridge_results.items())[:2])

        Nsignals = len(self.parentDV.df.columns)
        msg = f"Processed {Nsignals} signals!"
        msgBox = QMessageBox()
        msgBox.setWindowTitle("Batch processing done")
        msgBox.setText(msg)
        msgBox.exec()

    def get_thresh(self):

        """
        Reads the self.thresh_edit
        A Validator is set..
        """

        thresh_str = self.thresh_edit.text().replace(",", ".")
        try:
            thresh = float(thresh_str)
            if self.debug:
                print("thresh set to:", thresh)
            return thresh

        # empty line edit is interpreted as no thresholding required
        except ValueError:
            if self.debug:
                print("thresh ValueError", thresh_str)
            return 0

    def get_ridge_smooth(self):

        """
        Reads the self.smooth_edit
        A Validator is set..
        """

        rs = self.smooth_edit.text()
        try:
            rsmooth = int(rs)
            if self.debug:
                print("rsmooth set to:", rs)

            # make an odd window length
            if rsmooth == 0:
                return None
            elif rsmooth < 5:
                return 5
            elif rsmooth > 5 and rsmooth % 2 == 0:
                return rsmooth + 1
            else:
                return rsmooth

        # empty line edit is interpreted as no ridge smoothing required
        except ValueError:
            if self.debug:
                print("No rsmooth", rs)
            return None

    def get_OutPath(self):

        """
        Reads the self.OutPath_edit
        There is no validator but an os.path
        check is done!
        """

        path = self.OutPath_edit.text()

        if not os.path.isdir(path):

            msgBox = QMessageBox()
            msgBox.setWindowTitle("Invalid export path")
            msgBox.setText("Specified path is not a valid directory..")
            msgBox.exec()

            return None
        return path

    def select_export_dir(self):

        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.DirectoryOnly)
        dialog.setOption(QFileDialog.ShowDirsOnly, False)

        # retrieve or initialize directory path
        settings = QSettings()
        dir_path = settings.value("dir_name", expanduser("~"))

        dir_name = dialog.getExistingDirectory(
            self, "Select a folder to save the results", dir_path
        )

        # dialog cancelled
        if not dir_name:
            return

        if self.debug:
            print("Batch output name:", dir_name)

        self.OutPath_edit.setText(dir_name)

    def do_the_loop(self):

        """
        Uses the explicitly parsed self.wlet_pars
        to control signal analysis settings.

        Takes general analysis Parameters

        self.parentDV.dt
        self.parentDV.time_unit

        and the DataFrame

        self.parentDV.df

        from the parent DataViewer.
        Reads additional settings from this Batch Process Window.

        """

        EmptyRidge = 0

        if self.export_options.isChecked():
            OutPath = self.get_OutPath()
            if OutPath is None:
                return

        periods = np.linspace(
            self.wlet_pars["Tmin"], self.wlet_pars["Tmax"], self.wlet_pars["step_num"]
        )

        # retrieve batch settings
        power_thresh = self.get_thresh()
        rsmooth = self.get_ridge_smooth()

        # results get stored here
        ridge_results = {}
        df_fouriers = pd.DataFrame(index=periods)
        df_fouriers.index.name = "period"

        for i, signal_id in enumerate(self.parentDV.df):

            # log to terminal
            print(f"processing {signal_id}..")

            # sets parentDV.raw_signal and parentDV.tvec
            succ = self.parentDV.vector_prep(signal_id)
            # ui silently passes over..
            if not succ:
                print(f"Warning, can't process signal {signal_id}..")
                continue

            # detrend?!
            if self.parentDV.cb_use_detrended.isChecked():
                trend = self.parentDV.calc_trend()
                signal = self.parentDV.raw_signal - trend
            else:
                signal = self.parentDV.raw_signal

            # amplitude normalization?
            if self.parentDV.cb_use_envelope.isChecked():
                if self.debug:
                    print("Calculating envelope with L=", self.wlet_pars["window_size"])
                signal = pyboat.normalize_with_envelope(
                    signal, self.wlet_pars["window_size"], self.parentDV.dt
                )

            # compute the spectrum
            modulus, wlet = pyboat.compute_spectrum(signal, self.parentDV.dt, periods)
            # get maximum ridge
            ridge = pyboat.get_maxRidge_ys(modulus)
            # generate time vector
            tvec = np.arange(len(signal)) * self.parentDV.dt
            # evaluate along the ridge
            ridge_data = pyboat.eval_ridge(
                ridge,
                wlet,
                signal,
                periods,
                tvec,
                power_thresh,
                smoothing_wsize=rsmooth,
            )

            # from ridge thresholding..
            if ridge_data.empty:
                EmptyRidge += 1
            else:
                ridge_results[signal_id] = ridge_data

            # time average the spectrum, all have shape len(periods)!
            averaged_Wspec = np.mean(modulus, axis=1)
            df_fouriers[signal_id] = averaged_Wspec

            # -- Save out individual results --
            settings = QSettings()
            float_format = settings.value("float_format", "%.3f")
            graphics_format = settings.value("graphics_format", "png")

            exbox_checked = self.export_options.isChecked()

            if exbox_checked and self.cb_filtered_sigs.isChecked():

                signal_df = pd.DataFrame()
                signal_df["signal"] = signal
                signal_df.index = tvec
                signal_df.index.name = "time"

                fname = os.path.join(OutPath, f"{signal_id}_filtered.csv")
                if self.debug:
                    print(f"Saving filtered signal to {fname}")
                signal_df.to_csv(
                    fname, sep=",", float_format=float_format, index=True, header=True
                )

            if exbox_checked and self.cb_specs.isChecked():

                # plot spectrum and ridge
                ax_sig, ax_spec = pl.mk_signal_modulus_ax(self.parentDV.time_unit)
                pl.plot_signal_modulus(
                    (ax_sig, ax_spec),
                    tvec,
                    signal,
                    modulus,
                    periods,
                    p_max=self.wlet_pars["pow_max"],
                )
                pl.draw_Wavelet_ridge(ax_spec, ridge_data)
                plt.tight_layout()
                fname = os.path.join(OutPath, f"{signal_id}_wspec.{graphics_format}")
                if self.debug:
                    print(f"Plotting and saving spectrum {signal_id} to {fname}")
                plt.savefig(fname, dpi=DPI)
                plt.close()

            if exbox_checked and self.cb_specs_noridge.isChecked():

                # plot spectrum without ridge
                ax_sig, ax_spec = pl.mk_signal_modulus_ax(self.parentDV.time_unit)
                pl.plot_signal_modulus(
                    (ax_sig, ax_spec),
                    tvec,
                    signal,
                    modulus,
                    periods,
                    p_max=self.wlet_pars["pow_max"],
                )
                plt.tight_layout()
                fname = os.path.join(OutPath, f"{signal_id}_wspecNR.{graphics_format}")
                if self.debug:
                    print(f"Plotting and saving spectrum {signal_id} to {fname}")
                plt.savefig(fname, dpi=DPI)
                plt.close()

            if (
                exbox_checked
                and self.cb_readout_plots.isChecked()
                and not ridge_data.empty
            ):

                pl.plot_readout(ridge_data)
                fname = os.path.join(OutPath, f"{signal_id}_readout.{graphics_format}")
                if self.debug:
                    print(f"Plotting and saving {signal_id} to {fname}")
                plt.savefig(fname, dpi=DPI)
                plt.close()

            if exbox_checked and self.cb_readout.isChecked() and not ridge_data.empty:

                fname = os.path.join(OutPath, f"{signal_id}_readout.csv")
                if self.debug:
                    print(f"Saving ridge readout to {fname}")
                ridge_data.to_csv(
                    fname, sep=",", float_format=float_format, index=False
                )

            self.progress.setValue(i)

        if EmptyRidge > 0:

            msg = f"{EmptyRidge} ridge readouts entirely below threshold.."
            msgBox = QMessageBox()
            msgBox.setWindowTitle("Discarded Ridges")
            msgBox.setText(msg)
            msgBox.exec()

        return ridge_results, df_fouriers


class PowerDistributionWindow(QWidget):
    def __init__(self, powers, dataset_name, parent=None):
        super().__init__()

        # --- calculate average powers ------------------
        self.powers = powers
        # -------------------------------------------------

        self.initUI(dataset_name)

    def initUI(self, dataset_name):

        self.setWindowTitle(f"Ridge Power Distribution - {dataset_name}")
        self.setGeometry(410, 220, 550, 400)

        main_frame = QWidget()
        pCanvas = mkGenericCanvas()
        pCanvas.setParent(main_frame)
        ntb = NavigationToolbar(pCanvas, main_frame)

        # plot it
        pCanvas.fig.clf()
        pl.power_distribution(self.powers, fig=pCanvas.fig)
        pCanvas.fig.subplots_adjust(left=0.15, bottom=0.17)

        main_layout = QGridLayout()
        main_layout.addWidget(pCanvas, 0, 0, 9, 1)
        main_layout.addWidget(ntb, 10, 0, 1, 1)

        self.setLayout(main_layout)
        self.show()


class EnsembleDynamicsWindow(QWidget):
    def __init__(self, ensemble_results, dt, time_unit, dataset_name=""):
        super().__init__()

        self.time_unit = time_unit
        self.dt = dt
        # period, amplitude and phase
        self.results = ensemble_results

        self.initUI(dataset_name)

    def initUI(self, dataset_name):

        self.setWindowTitle(f"Ensemble Dynamics - {dataset_name}")
        self.setGeometry(210, 80, 700, 480)

        main_frame = QWidget()
        Canvas = mkGenericCanvas()
        Canvas.setParent(main_frame)
        ntb = NavigationToolbar(Canvas, main_frame)

        Canvas.fig.clf()
        pl.ensemble_dynamics(
            *self.results, dt=self.dt, time_unit=self.time_unit, fig=Canvas.fig
        )
        Canvas.fig.subplots_adjust(
            wspace=0.3, left=0.1, top=0.98, right=0.95, bottom=0.15
        )
        main_layout = QGridLayout()
        main_layout.addWidget(Canvas, 0, 0, 9, 1)
        main_layout.addWidget(ntb, 10, 0, 1, 1)

        self.setLayout(main_layout)
        self.show()


class FourierDistributionWindow(QWidget):
    def __init__(self, df_fouriers, time_unit, dataset_name=""):
        super().__init__()

        self.time_unit = time_unit
        self.df_fouriers = df_fouriers

        self.initUI(dataset_name)

    def initUI(self, dataset_name):

        self.setWindowTitle(f"Fourier Power Distribution - {dataset_name}")
        self.setGeometry(510, 330, 550, 400)

        main_frame = QWidget()
        Canvas = mkGenericCanvas()
        Canvas.setParent(main_frame)
        ntb = NavigationToolbar(Canvas, main_frame)

        Canvas.fig.clf()
        pl.Fourier_distribution(
            self.df_fouriers, time_unit=self.time_unit, fig=Canvas.fig
        )

        Canvas.fig.subplots_adjust(
            wspace=0.3, left=0.15, top=0.98, right=0.95, bottom=0.15
        )
        main_layout = QGridLayout()
        main_layout.addWidget(Canvas, 0, 0, 9, 1)
        main_layout.addWidget(ntb, 10, 0, 1, 1)

        self.setLayout(main_layout)
        self.show()



'''
This script shows how to estimate an empirical background from
an ensemble of 'non-oscillatory' signals, and how to apply this
estimated background to the analysis.
'''
import numpy as np
import pandas as pd
from pyboat import WAnalyzer, ssg
from pyboat import plotting as pl
from pyboat.core import ar1_powerspec
import matplotlib.pyplot as ppl

# set up analyzing instance
periods = np.linspace(5, 80, 100)
dt = 1
wAn = WAnalyzer(periods, dt, p_max=20)

# create an ensemble of short AR(1) realizations
# in a real life scenario this would be the 'non-oscillatory'
# test signals
alpha = 0.7
signals = [ssg.ar1_sim(alpha, Nt=120) for i in range(100)]

# store the individual time averaged Wavelet spectra
df_fouriers = pd.DataFrame(index = wAn.periods)

# the individual Fourier estimates
for i,sig in enumerate(signals):

    wAn.compute_spectrum(sig, do_plot=False)
    df_fouriers[i] = wAn.get_averaged_spectrum()

# plot the Fourier spectra distribution
pl.Fourier_distribution(df_fouriers)

# compare to the theoretical spectrum,
# note the finite size effects suppressing the power for higher periods!
theo_bg = ar1_powerspec(alpha, wAn.periods, wAn.dt)
ppl.plot(wAn.periods, theo_bg, 'k--', label='theoretical')
ppl.legend()

# now we take the median of the Fourier power
# distribution as expected empirical background
emp_bg = df_fouriers.median(axis=1)

# note that we have have persistent and consistent periods
print(len(emp_bg.index) == len(wAn.periods))

# let's create a very noisy signal with the same
# background process as noise mixed in 
test_signal = ssg.create_noisy_chirp(T1=70,
                                     T2=40,
                                     Nt=500,
                                     alpha=alpha,
                                     eps=1.) # SNR is 1!

# plain wavelet analysis
wAn.compute_spectrum(test_signal)

# let's draw the confidence intervals from the theoretical spectrum
wAn.draw_confidence_from_bg(theo_bg, colors='black')

# now let's use the 'empirical' one, taking into
# account the finite size effect
wAn.draw_confidence_from_bg(emp_bg, colors='orange')

# finally let's get the ridge within the CI regions
# from the empirical background
ridge_data = wAn.get_sign_maxRidge(emp_bg)
wAn.draw_Ridge()



import matplotlib.pyplot as ppl
import numpy as np

from pyboat import WAnalyzer, ssg

ppl.ion()

# --- set basic parameters and initialize the Analyzer---

dt = 2 # the sampling interval, 2s
periods = np.linspace(6, 90, 150) # period range, 6s to 90s
wAn = WAnalyzer(periods, dt, time_unit_label='s')

# --- create a synthetic signal ---

eps = 0.5 # noise intensity
alpha = 0.4 # AR1 parameter, set to 0 for white noise
Nt = 220 # number of samples

# oscillatory signal which sweeps from 30s to 50s
signal1 = ssg.create_noisy_chirp(T1 = 30 / dt, T2 = 50 / dt, Nt = Nt, eps = eps, alpha = alpha)

# add slower oscillatory trend with a period of 70s
signal2 = ssg.create_chirp(T1 = 70 / dt, T2 = 70 / dt, Nt = Nt)

# add exponential decay
syn_env = ssg.create_exp_envelope(tau = 0.65 * Nt, Nt = Nt)

# linear superposition
signal = syn_env * (signal1 + 2. * signal2)

# --- Filtering ---

# calculate the trend with a 60s cutoff
trend = wAn.sinc_smooth(signal, T_c=60)

# detrending here is then just a subtraction
detrended_signal = signal - trend

# normalize the amplitude with a sliding window of 70s
norm_signal = wAn.normalize_amplitude(detrended_signal, window_size=70)

# plot signal and trend, arguments can be any which ppl.plot(...) accepts
wAn.plot_signal(signal, label='Raw signal', color='red', alpha=0.5)
wAn.plot_trend(trend, label='Trend with $T_c$=60s')

# make a new figure to show original signal and detrended + normalized
wAn.plot_signal(signal, num=2, label='Raw signal', color='red', alpha=0.5)
wAn.plot_signal(norm_signal, label='Detrended + normalized', alpha=0.8, marker='.')

# --- Wavelet Transforms ---

# compute and plot the spectrum without detrending
wAn.compute_spectrum(signal)
wAn.ax_spec_signal.set_title('Raw input')
# compute and plot the spectrum of the detrended signal
# the low frequencies are gone!
wAn.compute_spectrum(detrended_signal)
wAn.ax_spec_signal.set_title('Sinc detrended')

# compute and plot the spectrum of the detrended
# and normalized signal, this function also returns
# the Wavelet transformation results directly
modulus, transform = wAn.compute_spectrum(norm_signal)
wAn.ax_spec_signal.set_title('Sinc detrended + ampl. normalized')

# zoom into the spec 
# wAn.ax_spec.set_ylim((20,60))

# change x-ticks
wAn.ax_spec.set_xticks([0, 50, 100, 150, 200, 250, 300, 350])
# also vertical grid
wAn.ax_spec.grid(axis='both', color='white', alpha=0.4)

# --- Ridge Evaluation and Readout ---

# get the ridge of the last analysis
wAn.get_maxRidge(power_thresh = 10, smoothing_wsize=20)
wAn.draw_Ridge()
# ppl.savefig('detr_signal_spec.png')

wAn.plot_readout(draw_coi=True)
# ppl.savefig('detr_signal_readout.png')

rd = wAn.ridge_data # this is a pandas DataFrame holding the ridge results
print(rd)



from PyQt5.QtWidgets import (
    QCheckBox,
    QMainWindow,
    QLabel,
    QLineEdit,
    QPushButton,
    QMessageBox,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QGroupBox,
    QFormLayout,
    QGridLayout,
    QTabWidget,
    QSpacerItem,
)

from PyQt5.QtGui import QDoubleValidator, QIntValidator
from PyQt5.QtCore import Qt

from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar

import pyboat
from pyboat.ui.util import MessageWindow, posfloatV, posintV, floatV, set_max_width
from pyboat.ui.analysis import mkTimeSeriesCanvas, FourierAnalyzer, WaveletAnalyzer
from pyboat import plotting as pl
from pyboat import ssg  # the synthetic signal generator
import numpy as np

# --- monkey patch label sizes good for ui ---
pl.tick_label_size = 12
pl.label_size = 14


class SynthSignalGen(QMainWindow):

    """
    This is basically a clone of the
    DataViewer, but instead of the
    imported data, controls for
    generating a synthetic signal
    are provided
    """

    def __init__(self, parent, debug=False):
        super().__init__(parent=parent)

        self.anaWindows = {}
        self.w_position = 0  # analysis window position offset

        self.debug = debug

        self.raw_signal = None
        self.Nt = 500  # needs to be set here for initial plot..
        self.dt = 1  # .. to create the default signal
        self.tvec = None  # gets initialized by vector_prep
        self.time_unit = None  # gets initialized by qset_time_unit

        # get updated with dt in -> qset_dt
        self.periodV = QDoubleValidator(bottom=1e-16, top=1e16)
        self.envelopeV = QDoubleValidator(bottom=3, top=9999999)

        self.initUI()

    # ===============UI=======================================

    def initUI(self):

        self.setWindowTitle(f"Synthetic Signal Generator")
        self.setGeometry(80, 300, 900, 650)

        main_widget = QWidget()
        self.statusBar()

        self.tsCanvas = mkTimeSeriesCanvas()
        main_frame = QWidget()
        self.tsCanvas.setParent(main_frame)
        ntb = NavigationToolbar(self.tsCanvas, main_frame)  # full toolbar

        main_layout_v = QVBoxLayout()  # The whole Layout

        # --- the synthesizer controls ---

        # width of the input fields
        iwidth = 50

        dt_label = QLabel("Sampling Interval:")
        dt_edit = QLineEdit()
        set_max_width(dt_edit, iwidth)
        dt_edit.setValidator(posfloatV)
        dt_edit.textChanged[str].connect(self.qset_dt)

        unit_label = QLabel("Time Unit:")
        unit_edit = QLineEdit(self)
        set_max_width(unit_edit, iwidth)
        unit_edit.textChanged[str].connect(self.qset_time_unit)
        unit_edit.insert("min")  # standard time unit is minutes

        Nt_label = QLabel("# Samples")
        self.Nt_edit = QLineEdit()
        self.Nt_edit.setStatusTip(
            "Number of data points, minimum is 10, maximum is 25 000!"
        )
        set_max_width(self.Nt_edit, iwidth)
        self.Nt_edit.setValidator(QIntValidator(bottom=10, top=25000))

        # --- the basic settings box ---
        basics_box = QGroupBox("Basics")
        basics_box_layout = QVBoxLayout()
        basics_box_layout.setSpacing(2)
        basics_box.setLayout(basics_box_layout)

        basics_box_layout.addWidget(Nt_label)
        basics_box_layout.addWidget(self.Nt_edit)
        basics_box_layout.addWidget(dt_label)
        basics_box_layout.addWidget(dt_edit)
        basics_box_layout.addWidget(unit_label)
        basics_box_layout.addWidget(unit_edit)

        basics_box_layout.addStretch(0)

        # --- chirp 1 ---

        T11_label = QLabel("Initial Period")
        self.T11_edit = QLineEdit()
        self.T11_edit.setStatusTip("Period at the beginning of the signal")
        set_max_width(self.T11_edit, iwidth)
        self.T11_edit.setValidator(posfloatV)
        self.T11_edit.insert(str(50))  # initial period of chirp 1

        T12_label = QLabel("Final Period")
        self.T12_edit = QLineEdit()
        self.T12_edit.setStatusTip("Period at the end of the signal")
        set_max_width(self.T12_edit, iwidth)
        self.T12_edit.setValidator(posfloatV)
        self.T12_edit.insert(str(150))  # initial period of chirp 1

        A1_label = QLabel("Amplitude")
        self.A1_edit = QLineEdit()
        self.A1_edit.setStatusTip("The amplitude :)")
        set_max_width(self.A1_edit, iwidth)
        self.A1_edit.setValidator(posfloatV)
        self.A1_edit.insert(str(1))  # initial amplitude

        # --- the chirp 1 box ---
        self.chirp1_box = QGroupBox("Oscillator I")
        self.chirp1_box.setCheckable(True)
        chirp1_box_layout = QVBoxLayout()
        chirp1_box_layout.setSpacing(2)
        self.chirp1_box.setLayout(chirp1_box_layout)

        chirp1_box_layout.addWidget(T11_label)
        chirp1_box_layout.addWidget(self.T11_edit)
        chirp1_box_layout.addWidget(T12_label)
        chirp1_box_layout.addWidget(self.T12_edit)
        chirp1_box_layout.addWidget(A1_label)
        chirp1_box_layout.addWidget(self.A1_edit)
        chirp1_box_layout.addStretch(0)

        # --- chirp 2 ---
        # can be used to simulate a trend :)

        T21_label = QLabel("Initial Period")
        self.T21_edit = QLineEdit()
        self.T21_edit.setStatusTip("Period at the beginning of the signal")
        set_max_width(self.T21_edit, iwidth)
        self.T21_edit.setValidator(posfloatV)
        self.T21_edit.insert(str(1000))  # initial period of chirp 1

        T22_label = QLabel("Final Period")
        self.T22_edit = QLineEdit()
        self.T22_edit.setStatusTip("Period at the end of the signal")
        set_max_width(self.T22_edit, iwidth)
        self.T22_edit.setValidator(posfloatV)
        self.T22_edit.insert(str(1000))  # initial period of chirp 1

        A2_label = QLabel("Amplitude")
        self.A2_edit = QLineEdit()
        self.A2_edit.setStatusTip("Can be negative to induce different trends..")
        set_max_width(self.A2_edit, iwidth)
        self.A2_edit.setValidator(floatV)
        self.A2_edit.insert(str(2))  # initial amplitude

        # --- the chirp 2 box ---
        self.chirp2_box = QGroupBox("Oscillator II")
        self.chirp2_box.setCheckable(True)
        chirp2_box_layout = QVBoxLayout()
        self.chirp2_box.setLayout(chirp2_box_layout)
        chirp2_box_layout.setSpacing(2)

        chirp2_box_layout.addWidget(T21_label)
        chirp2_box_layout.addWidget(self.T21_edit)
        chirp2_box_layout.addWidget(T22_label)
        chirp2_box_layout.addWidget(self.T22_edit)
        chirp2_box_layout.addWidget(A2_label)
        chirp2_box_layout.addWidget(self.A2_edit)
        chirp2_box_layout.addStretch(0)

        # --- the AR1 box ---
        self.noise_box = QGroupBox("Noise")
        self.noise_box.setStatusTip("Adds colored AR(1) noise to the signal")
        self.noise_box.setCheckable(True)
        self.noise_box.setChecked(False)
        noise_box_layout = QVBoxLayout()
        self.noise_box.setLayout(noise_box_layout)

        alpha_label = QLabel("AR(1) parameter")
        self.alpha_edit = QLineEdit(parent=self.noise_box)
        self.alpha_edit.setStatusTip("0 is white noise, must be smaller than 1!")
        set_max_width(self.alpha_edit, iwidth)
        # does not work as expected :/
        V1 = QDoubleValidator(bottom=0, top=0.99)
        # V1.setNotation(QDoubleValidator.StandardNotation)
        # V1.setDecimals(5)
        # V1.setRange(0.0, 0.99)
        self.alpha_edit.setValidator(V1)
        self.alpha_edit.insert("0.2")  # initial AR(1)

        d_label = QLabel("Noise Strength")
        self.d_edit = QLineEdit(parent=self.noise_box)
        self.d_edit.setStatusTip("Relative to oscillator amplitudes..")
        set_max_width(self.d_edit, iwidth)
        self.d_edit.setValidator(QDoubleValidator(bottom=0, top=999999))
        self.d_edit.insert("0.5")  # initial noise strength

        noise_box_layout.addWidget(alpha_label)
        noise_box_layout.addWidget(self.alpha_edit)
        noise_box_layout.addWidget(d_label)
        noise_box_layout.addWidget(self.d_edit)
        noise_box_layout.addStretch(0)

        # --- Amplitude envelope ---

        tau_label = QLabel("Decay Time")
        self.tau_edit = QLineEdit()
        self.tau_edit.setStatusTip(
            "Time after which the signal decayed to around a third of the initial amplitude"
        )
        set_max_width(self.tau_edit, iwidth)
        self.tau_edit.setValidator(posfloatV)
        self.tau_edit.insert("500")  # initial decay constant

        # --- the Envelope box ---
        self.env_box = QGroupBox("Exponential Envelope")
        self.env_box.setCheckable(True)
        env_box_layout = QVBoxLayout()
        env_box_layout.setSpacing(2)
        self.env_box.setLayout(env_box_layout)

        env_box_layout.addWidget(tau_label)
        env_box_layout.addWidget(self.tau_edit)
        env_box_layout.addStretch(0)

        # --- the create signal button
        ssgButton = QPushButton("Synthesize Signal", self)
        ssgButton.clicked.connect(self.create_signal)
        ssgButton.setStatusTip(
            "Click again with same settings for different noise realizations"
        )
        ssgButton.setStyleSheet("background-color: orange")

        ssgButton_box = QWidget()
        ssgButton_box_layout = QHBoxLayout()
        ssgButton_box.setLayout(ssgButton_box_layout)
        # ssgButton_box_layout.addStretch(0)
        ssgButton_box_layout.addWidget(ssgButton)
        # ssgButton_box_layout.addStretch(0)

        control_grid = QWidget()
        control_grid_layout = QGridLayout()
        control_grid.setLayout(control_grid_layout)

        control_grid_layout.addWidget(basics_box, 0, 0, 4, 1)
        control_grid_layout.addWidget(self.chirp1_box, 0, 1, 4, 1)
        control_grid_layout.addWidget(self.chirp2_box, 0, 2, 4, 1)
        control_grid_layout.addWidget(self.noise_box, 0, 3, 4, 1)
        control_grid_layout.addWidget(self.env_box, 0, 4, 1, 1)
        control_grid_layout.addItem(QSpacerItem(2, 15), 1, 4)
        control_grid_layout.addWidget(ssgButton_box, 2, 4)
        control_grid_layout.addItem(QSpacerItem(2, 15), 3, 4)

        controls = QWidget()
        controls_layout = QVBoxLayout()
        controls.setLayout(controls_layout)
        controls_layout.addWidget(control_grid)

        main_layout_v.addWidget(controls)

        ## detrending parameter

        self.T_c_edit = QLineEdit()
        self.T_c_edit.setMaximumWidth(70)
        self.T_c_edit.setValidator(posfloatV)

        sinc_options_box = QGroupBox("Sinc Detrending")
        sinc_options_layout = QGridLayout()
        sinc_options_layout.addWidget(QLabel("Cut-off Period:"), 0, 0)
        sinc_options_layout.addWidget(self.T_c_edit, 0, 1)
        sinc_options_box.setLayout(sinc_options_layout)

        ## Amplitude envelope parameter
        self.L_edit = QLineEdit()
        self.L_edit.setMaximumWidth(70)
        self.L_edit.setValidator(self.envelopeV)

        envelope_options_box = QGroupBox("Amplitude Envelope")
        envelope_options_layout = QGridLayout()
        envelope_options_layout.addWidget(QLabel("Window Size:"), 0, 0)
        envelope_options_layout.addWidget(self.L_edit, 0, 1)
        envelope_options_box.setLayout(envelope_options_layout)

        # plot options box
        plot_options_box = QGroupBox("Plotting Options")
        plot_options_layout = QGridLayout()

        self.cb_raw = QCheckBox("Raw Signal", self)
        self.cb_trend = QCheckBox("Trend", self)
        self.cb_detrend = QCheckBox("Detrended Signal", self)
        self.cb_envelope = QCheckBox("Envelope", self)

        plotButton = QPushButton("Refresh Plot", self)
        plotButton.clicked.connect(self.doPlot)

        ## checkbox layout
        plot_options_layout.addWidget(self.cb_raw, 0, 0)
        plot_options_layout.addWidget(self.cb_trend, 0, 1)
        plot_options_layout.addWidget(self.cb_detrend, 1, 0)
        plot_options_layout.addWidget(self.cb_envelope, 1, 1)
        plot_options_layout.addWidget(plotButton, 2, 0)
        plot_options_box.setLayout(plot_options_layout)

        ## checkbox signal set and change
        self.cb_raw.toggle()
        self.cb_trend.toggle()

        self.cb_raw.stateChanged.connect(self.toggle_raw)
        self.cb_trend.stateChanged.connect(self.toggle_trend)
        self.cb_detrend.stateChanged.connect(self.toggle_trend)
        self.cb_envelope.stateChanged.connect(self.toggle_envelope)

        # Ploting box/Canvas area
        plot_box = QGroupBox("Signal and Trend")
        plot_layout = QVBoxLayout()
        plot_layout.addWidget(self.tsCanvas)
        plot_layout.addWidget(ntb)
        plot_box.setLayout(plot_layout)

        # Analyzer box with tabs
        ana_widget = QGroupBox("Analysis")
        ana_box = QVBoxLayout()

        ## Initialize tab scresen
        tabs = QTabWidget()
        tab1 = QWidget()
        tab2 = QWidget()

        ## Add tabs
        tabs.addTab(tab1, "Wavelet Analysis")
        tabs.addTab(tab2, "Fourier Transform")

        ## Create first tab
        tab1.parameter_box = QFormLayout()

        ## for wavlet params, button, etc.
        self.T_min = QLineEdit()
        self.T_min.setStatusTip("This is the lower period limit")
        self.step_num = QLineEdit()
        self.step_num.insert("200")
        self.step_num.setStatusTip("Increase this number for more resolution")
        self.T_max = QLineEdit()
        self.T_max.setStatusTip("This is the upper period limit")

        self.p_max = QLineEdit()
        self.p_max.setStatusTip(
            "Enter a fixed value for all signals or leave blank for automatic scaling"
        )

        # self.p_max.insert(str(20)) # leave blank

        T_min_lab = QLabel("Lowest period")
        step_lab = QLabel("Number of periods")
        T_max_lab = QLabel("Highest  period")
        p_max_lab = QLabel("Expected maximal power")

        T_min_lab.setWordWrap(True)
        step_lab.setWordWrap(True)
        T_max_lab.setWordWrap(True)
        p_max_lab.setWordWrap(True)

        wletButton = QPushButton("Analyze Signal", self)
        wletButton.clicked.connect(self.run_wavelet_ana)
        wletButton.setStatusTip("Start the wavelet analysis!")
        wletButton.setStyleSheet("background-color: lightblue")

        ## add  button to layout
        wlet_button_layout_h = QHBoxLayout()

        wlet_button_layout_h.addStretch(0)
        wlet_button_layout_h.addWidget(wletButton)
        wlet_button_layout_h.addStretch(0)

        self.cb_use_detrended = QCheckBox("Use Detrended Signal", self)

        # self.cb_use_detrended.stateChanged.connect(self.toggle_use)
        self.cb_use_detrended.setChecked(True)  # detrend by default

        self.cb_use_envelope = QCheckBox("Normalize with Envelope", self)
        self.cb_use_envelope.setChecked(False)  # no envelope by default

        ## Add Wavelet analyzer options to tab1.parameter_box layout

        tab1.parameter_box.addRow(T_min_lab, self.T_min)
        tab1.parameter_box.addRow(T_max_lab, self.T_max)
        tab1.parameter_box.addRow(step_lab, self.step_num)
        tab1.parameter_box.addRow(p_max_lab, self.p_max)
        tab1.parameter_box.addRow(self.cb_use_detrended)
        tab1.parameter_box.addRow(self.cb_use_envelope)
        tab1.parameter_box.addRow(wlet_button_layout_h)

        tab1.setLayout(tab1.parameter_box)

        # fourier button
        fButton = QPushButton("Analyze Signal", self)
        ## add  button to layout
        f_button_layout_h = QHBoxLayout()
        fButton.clicked.connect(self.run_fourier_ana)
        f_button_layout_h.addStretch(0)
        f_button_layout_h.addWidget(fButton)

        # fourier detrended switch
        self.cb_use_detrended2 = QCheckBox("Use Detrended Signal", self)
        self.cb_use_detrended2.setChecked(True)  # detrend by default

        self.cb_use_envelope2 = QCheckBox("Normalize with Envelope", self)
        self.cb_use_envelope2.setChecked(False)

        # fourier period or frequency view
        self.cb_FourierT = QCheckBox("Show Frequencies", self)
        self.cb_FourierT.setChecked(False)  # show periods per default

        ## Create second tab
        tab2.parameter_box = QFormLayout()
        # tab2.parameter_box.addRow(T_min_lab,self.T_min)
        # tab2.parameter_box.addRow(T_max_lab,self.T_max)
        tab2.parameter_box.addRow(self.cb_use_detrended2)
        tab2.parameter_box.addRow(self.cb_use_envelope2)
        tab2.parameter_box.addRow(self.cb_FourierT)
        tab2.parameter_box.addRow(f_button_layout_h)
        tab2.setLayout(tab2.parameter_box)

        # Add tabs to Vbox
        ana_box.addWidget(tabs)
        # set layout of ana_widget (will be added to options layout)
        # as ana_box (containing actual layout)
        ana_widget.setLayout(ana_box)

        # Fix size of table_widget containing parameter boxes
        # -> it's all done via column stretches of
        # the GridLayout below
        # size_pol= QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        # ana_widget.setSizePolicy(size_pol)

        # ==========Plot and Options Layout=======================================
        # Merge Plotting canvas and options
        plot_and_options = QWidget()
        layout = QGridLayout()
        plot_and_options.setLayout(layout)

        layout.addWidget(plot_box, 0, 0, 4, 1)
        layout.addWidget(sinc_options_box, 0, 5, 1, 1)
        layout.addWidget(envelope_options_box, 0, 6, 1, 1)
        layout.addWidget(plot_options_box, 1, 5, 1, 2)
        layout.addWidget(ana_widget, 2, 5, 2, 2)

        # plotting-canvas column stretch <-> 1st (0th) column
        layout.setColumnStretch(0, 1)  # plot should stretch
        layout.setColumnMinimumWidth(0, 360)  # plot should not get too small

        layout.setColumnStretch(1, 0)  # options shouldn't stretch
        layout.setColumnStretch(2, 0)  # options shouldn't stretch

        # ==========Main Layout=======================================
        main_layout_v.addWidget(plot_and_options)  # is a VBox
        main_layout_v.setStretch(0, 0)  # controls shouldn't stretch
        main_layout_v.setStretch(1, 5)  # plot should stretch

        # --- initialize parameter fields only now ---

        # this will trigger setting initial periods
        self.Nt_edit.insert(str(self.Nt))  # initial signal length
        dt_edit.insert(str(self.dt))  # initial sampling interval is 1

        # create the default signal
        self.create_signal()

        main_widget.setLayout(main_layout_v)
        self.setCentralWidget(main_widget)
        self.show()

    # probably all the toggle state variables are not needed -> read out checkboxes directly
    def toggle_raw(self, state):
        if state == Qt.Checked:
            self.plot_raw = True
            # untoggle the detrended cb
            self.cb_detrend.setChecked(False)
        else:
            self.plot_raw = False

        if self.raw_signal is not None:
            self.doPlot()

    def toggle_trend(self, state):

        if self.debug:
            print("new state:", self.cb_trend.isChecked(), state, Qt.Checked)

        # trying to plot the trend
        if state == Qt.Checked:
            T_c = self.get_T_c(self.T_c_edit)
            if not T_c:
                if self.cb_trend.isChecked():
                    self.cb_trend.setChecked(False)
                if self.cb_detrend.isChecked():
                    self.cb_detrend.setChecked(False)
                self.cb_use_detrended.setChecked(False)
                return

            # don't plot raw and detrended together (trend is ok)
            if self.cb_detrend.isChecked():
                self.cb_raw.setChecked(False)

        # signal selected?
        if np.any(self.raw_signal):
            self.doPlot()

    def toggle_envelope(self, state):

        # trying to plot the envelope
        if state == Qt.Checked:
            L = self.get_L(self.L_edit)
            if not L:
                self.cb_envelope.setChecked(False)
                self.cb_use_envelope.setChecked(False)

        # signal selected?
        if np.any(self.raw_signal):
            self.doPlot()

    # connected to unit_edit
    def qset_time_unit(self, text):
        self.time_unit = text  # self.unit_edit.text()
        if self.debug:
            print("time unit changed to:", text)

    # connected to dt_edit
    def qset_dt(self, text):

        # checking the input is done automatically via .setValidator!
        # check,str_val,_ = posfloatV.validate(t,  0) # pos argument not used
        t = text.replace(",", ".")
        try:
            self.dt = float(t)
            self.set_initial_periods(force=True)
            self.set_initial_T_c(force=True)
            # update  Validators
            self.periodV = QDoubleValidator(bottom=2 * self.dt, top=1e16)
            self.envelopeV = QDoubleValidator(bottom=3 * self.dt, top=self.Nt * self.dt)

            # refresh plot if a is signal selected
            if self.raw_signal is not None:
                self.doPlot()

        # empty input, keeps old value!
        except ValueError:
            if self.debug:
                print("dt ValueError", text)
            pass

        if self.debug:
            print("dt set to:", self.dt)

    def get_T_c(self, T_c_edit):

        """
        Uses self.T_c_edit, argument just for clarity. Checks
        for empty input, this function only gets called when
        a detrending operation is requested. Hence, an empty
        QLineEdit will display a user warning and return nothing..
        """

        # value checking done by validator, accepts also comma '1,1' !
        tc = T_c_edit.text().replace(",", ".")
        try:
            T_c = float(tc)
            if self.debug:
                print("T_c set to:", T_c)
            return T_c

        # empty line edit
        except ValueError:
            msgBox = QMessageBox(parent=self)
            msgBox.setWindowTitle("Missing Parameter")
            msgBox.setText(
                "Detrending parameter not set,\n" + "specify a cut-off period!"
            )
            msgBox.exec()

            if self.debug:
                print("T_c ValueError", tc)
            return None

    def get_L(self, L_edit):

        # value checking done by validator, accepts also comma '1,1' !
        L = L_edit.text().replace(",", ".")
        try:
            L = int(L)

        # empty line edit
        except ValueError:
            msgBox = QMessageBox(parent=self)
            msgBox.setWindowTitle("Missing Parameter")
            msgBox.setText(
                "Envelope parameter not set,\n" + "specify a sliding window size!"
            )
            msgBox.exec()
            if self.debug:
                print("L ValueError", L)
            return None

        if self.debug:
            print("L set to:", L)
        return L

    def set_initial_periods(self, force=False):

        """
        rewrite value if force is True
        """

        if self.debug:
            print("set_initial_periods called")

        # check if a T_min was already entered
        # or rewrite is enforced
        if not bool(self.T_min.text()) or force:
            self.T_min.clear()
            self.T_min.insert(str(2 * self.dt))  # Nyquist

        if np.any(self.raw_signal):  # check if raw_signal already selected
            # check if a T_max was already entered
            if not bool(self.T_max.text()) or force:

                # default is 1/4 the observation time
                self.T_max.clear()
                T_max_ini = self.dt * 1 / 4 * self.Nt
                if self.dt > 0.1:
                    T_max_ini = int(T_max_ini)
                self.T_max.insert(str(T_max_ini))

    def set_initial_T_c(self, force=False):

        if self.debug:
            print("set_initial_T_c called")

        if np.any(self.raw_signal):  # check if raw_signal already selected
            if (
                not bool(self.T_c_edit.text()) or force
            ):  # check if a T_c was already entered
                # default is 1.5 * T_max -> 3/8 the observation time
                self.T_c_edit.clear()
                # this will trigger qset_T_c and updates the variable
                T_c_ini = self.dt * 3 / 8 * len(self.raw_signal)
                if self.dt > 0.1:
                    T_c_ini = int(T_c_ini)
                else:
                    T_c_ini = np.round(T_c_ini, 3)
                self.T_c_edit.insert(str(T_c_ini))

    def set_wlet_pars(self):

        """
        Retrieves and checks the set wavelet parameters
        of the 'Analysis' input box reading the following
        QLineEdits:

        self.Tmin
        self.Tmax
        self.step_num
        self.pmax

        Further the checkboxes regarding detrending and amplitude
        normalization are evaluated. And

        self.get_L()
        self.get_T_c()

        are called if needed. These respective parameters are set to False
        if the opereation in question is not requested.

        Returns
        -------

        wlet_pars : dictionary holding the retrieved parameters,
                    L and T_c are set to None if no amplitude
                    normalization or detrending operation should be done

        """

        wlet_pars = {}

        # period validator
        vali = self.periodV

        # -- read all the QLineEdits --

        text = self.T_min.text()
        T_min = text.replace(",", ".")
        check, _, _ = vali.validate(T_min, 0)
        if self.debug:
            print("Min periodValidator output:", check, "value:", T_min)
        if check == 0:
            self.OutOfBounds = MessageWindow(
                "Wavelet periods out of bounds!", "Value Error"
            )
            return False

        wlet_pars["T_min"] = float(T_min)

        step_num = self.step_num.text()
        check, _, _ = posintV.validate(step_num, 0)
        if self.debug:
            print("# Periods posintValidator:", check, "value:", step_num)
        if check == 0:
            self.OutOfBounds = MessageWindow(
                "The Number of periods must be a positive integer!", "Value Error"
            )
            return False

        wlet_pars["step_num"] = int(step_num)
        if int(step_num) > 1000:

            choice = QMessageBox.question(
                self,
                "Too much periods?: ",
                "High number of periods: Do you want to continue?",
                QMessageBox.Yes | QMessageBox.No,
            )
            if choice == QMessageBox.Yes:
                pass
            else:
                return False

        text = self.T_max.text()
        T_max = text.replace(",", ".")
        check, _, _ = vali.validate(T_max, 0)
        if self.debug:
            print("Max periodValidator output:", check)
            print(f"Max period value: {self.T_max.text()}")
        if check == 0:
            self.OutOfBounds = MessageWindow(
                "Wavelet highest period out of bounds!", "Value Error"
            )
            return False
        wlet_pars["T_max"] = float(T_max)

        text = self.p_max.text()
        p_max = text.replace(",", ".")
        check, _, _ = posfloatV.validate(p_max, 0)  # checks for positive float
        if check == 0:
            self.OutOfBounds = MessageWindow("Powers are positive!", "Value Error")
            return False

        # check for empty string:
        if p_max:
            wlet_pars["p_max"] = float(p_max)
        else:
            wlet_pars["p_max"] = None

        # -- the checkboxes --

        # detrend for the analysis?
        if self.cb_use_detrended.isChecked():
            T_c = self.get_T_c(self.T_c_edit)
            if T_c is None:
                return False  # abort settings
            wlet_pars["T_c"] = T_c
        else:
            # indicates no detrending requested
            wlet_pars["T_c"] = False

        # amplitude normalization is downstram of detrending!
        if self.cb_use_envelope.isChecked():
            L = self.get_L(self.L_edit)
            if L is None:
                return False  # abort settings
            wlet_pars["L"] = L
        else:
            # indicates no ampl. normalization
            wlet_pars["L"] = False

        # success!
        return wlet_pars

    def calc_trend(self):

        """ Uses maximal sinc window size """

        T_c = self.get_T_c(self.T_c_edit)
        if not T_c:
            return
        if self.debug:
            print("Calculating trend with T_c = ", T_c)

        trend = pyboat.sinc_smooth(raw_signal=self.raw_signal, T_c=T_c, dt=self.dt)
        return trend

    def calc_envelope(self):

        L = self.get_L(self.L_edit)
        if not L:
            return
        if self.debug:
            print("Calculating envelope with L = ", L)

        if L / self.dt < 4:
            self.OutOfBounds = MessageWindow(
                f"Minimum sliding window size is {4*self.dt}{self.time_unit} !",
                "Value Error",
            )
            L = None
            return

        if L / self.dt > self.raw_signal.size:
            maxL = self.raw_signal.size * self.dt
            self.OutOfBounds = MessageWindow(
                f"Maximum sliding window size is {maxL:.2f} {self.time_unit}!",
                "Value Error",
            )
            L = None
            return

        # cut of frequency set and detended plot activated?
        if self.cb_detrend.isChecked():

            trend = self.calc_trend()
            if trend is None:
                return
            signal = self.raw_signal - trend
        else:
            signal = self.raw_signal

        envelope = pyboat.sliding_window_amplitude(
            signal, window_size=L, dt=self.dt
        )

        return envelope

    def create_signal(self):

        """
        Retrieves all paramters from the synthesizer controls
        and calls the ssg. All line edits have validators set,
        however we have to check for intermediate empty inputs..

        This sets

        self.raw_signal

        according to the synthesizer controls.
        """

        # the signal components
        components = []
        weights = []

        # number of sample points
        if not self.Nt_edit.hasAcceptableInput():
            self.OutOfBounds = MessageWindow(
                "Minimum number of sample points is 10!", "Value Error"
            )
            return

        self.Nt = int(self.Nt_edit.text())
        if self.debug:
            print("Nt changed to:", self.Nt)

        self.set_initial_periods(force=False)
        self.set_initial_T_c(force=False)

        T_edits = [self.T11_edit, self.T12_edit, self.T21_edit, self.T22_edit]

        # check for valid inputs
        for T_e in T_edits:

            if self.debug:
                print(f"Is enabled: {T_e.isEnabled()}")
                print(f"Checking T_edits: {T_e.text()}")
                print(f"Validator output: {T_e.hasAcceptableInput()}")

            if not T_e.isEnabled():
                continue

            if not T_e.hasAcceptableInput():
                self.OutOfBounds = MessageWindow(
                    "All periods must be greater than 0!", "Value Error"
                )
                return

        # envelope before chirp creation
        if self.env_box.isChecked():

            if not self.tau_edit.hasAcceptableInput():
                self.OutOfBounds = MessageWindow(
                    "Missing envelope parameter!", "Value Error"
                )
                return

            tau = float(self.tau_edit.text()) / self.dt
            env = ssg.create_exp_envelope(tau, self.Nt)
            if self.debug:
                print(f"Creating the envelope with tau = {tau}")

        else:
            env = 1  # no envelope

        if self.chirp1_box.isChecked():

            if not self.A1_edit.hasAcceptableInput():
                self.OutOfBounds = MessageWindow(
                    "Set an amplitude for oscillator1!", "Value Error"
                )
                return

            # the periods
            T11 = float(self.T11_edit.text()) / self.dt
            T12 = float(self.T12_edit.text()) / self.dt
            A1 = float(self.A1_edit.text())
            chirp1 = ssg.create_chirp(T11, T12, self.Nt)
            components.append(env * chirp1)
            weights.append(A1)

        if self.chirp2_box.isChecked():

            if not self.A2_edit.hasAcceptableInput():
                self.OutOfBounds = MessageWindow(
                    "Set an amplitude for oscillator2!", "Value Error"
                )
                return

            T21 = float(self.T21_edit.text()) / self.dt
            T22 = float(self.T22_edit.text()) / self.dt
            A2 = float(self.A2_edit.text())
            chirp2 = ssg.create_chirp(T21, T22, self.Nt)
            components.append(env * chirp2)
            weights.append(A2)

        # noise
        if self.noise_box.isChecked():

            try:
                # QDoubleValidator is a screw up..
                alpha = float(self.alpha_edit.text())
            # empty input
            except ValueError:
                self.OutOfBounds = MessageWindow(
                    "Missing AR(1) alpha parameter!", "Value Error"
                )
                return

            if not 0 <= alpha < 1:
                self.OutOfBounds = MessageWindow(
                    "AR1 parameter must be between 0 and <1!", "Value Error"
                )
                return

            # Noise amplitude
            try:
                d = float(self.d_edit.text())

            except ValueError:
                self.OutOfBounds = MessageWindow(
                    "Missing Noise Strength parameter!", "Value Error"
                )
                return

            d = float(self.d_edit.text())
            noise = ssg.ar1_sim(alpha, self.Nt)
            components.append(noise)
            weights.append(d)

        if len(components) == 0:
            self.OutOfBounds = MessageWindow(
                "Activate at least one signal component!", "No Signal"
            )
            return

        signal = ssg.assemble_signal(components, weights)

        # ----------------------------------------
        self.raw_signal = signal
        self.tvec = self.dt * np.arange(self.Nt)
        # ---------------------------------------

        if self.debug:
            print("created synth. signal:", self.raw_signal[:10])

        # plot right away
        self.set_initial_periods()
        self.set_initial_T_c()
        self.doPlot()

    def doPlot(self):

        if self.raw_signal is None:
            self.NoSignal = MessageWindow("Please create a signal first!", "No Signal")

        if self.debug:
            print(
                "called Plotting [raw] [trend] [detrended] [envelope]",
                self.cb_raw.isChecked(),
                self.cb_trend.isChecked(),
                self.cb_detrend.isChecked(),
                self.cb_envelope.isChecked(),
            )

        # check if trend is needed
        if self.cb_trend.isChecked() or self.cb_detrend.isChecked():
            trend = self.calc_trend()
            if trend is None:
                return
        else:
            trend = None

        # envelope calculation
        if self.cb_envelope.isChecked():
            envelope = self.calc_envelope()
            if envelope is None:
                return

        else:
            envelope = None

        self.tsCanvas.fig1.clf()

        ax1 = pl.mk_signal_ax(self.time_unit, fig=self.tsCanvas.fig1)
        self.tsCanvas.fig1.add_axes(ax1)

        if self.debug:
            print(
                f"plotting signal and trend with {self.tvec[:10]}, {self.raw_signal[:10]}"
            )

        if self.cb_raw.isChecked():
            pl.draw_signal(ax1, time_vector=self.tvec, signal=self.raw_signal)

        if trend is not None and self.cb_trend.isChecked():
            pl.draw_trend(ax1, time_vector=self.tvec, trend=trend)

        if trend is not None and self.cb_detrend.isChecked():
            ax2 = pl.draw_detrended(
                ax1, time_vector=self.tvec, detrended=self.raw_signal - trend
            )
            ax2.legend(fontsize=pl.tick_label_size, loc="lower left")
        if envelope is not None and not self.cb_detrend.isChecked():
            pl.draw_envelope(ax1, time_vector=self.tvec, envelope=envelope)

        # plot on detrended axis
        if envelope is not None and self.cb_detrend.isChecked():
            pl.draw_envelope(ax2, time_vector=self.tvec, envelope=envelope)
            ax2.legend(fontsize=pl.tick_label_size, loc="lower left")
        self.tsCanvas.fig1.subplots_adjust(bottom=0.15, left=0.15, right=0.85)
        # add a simple legend
        ax1.legend(fontsize=pl.tick_label_size)

        self.tsCanvas.draw()
        self.tsCanvas.show()

    def run_wavelet_ana(self):

        """ run the Wavelet Analysis on the synthetic signal """

        if not np.any(self.raw_signal):
            self.NoSignalSelected = MessageWindow(
                "Please create a signal first!", "No Signal"
            )
            return False

        wlet_pars = self.set_wlet_pars()  # Error handling done there
        if not wlet_pars:
            if self.debug:
                print("Wavelet parameters could not be set!")
            return False

        if self.cb_use_detrended.isChecked():
            trend = self.calc_trend()
            signal = self.raw_signal - trend
        else:
            signal = self.raw_signal

        if self.cb_use_envelope.isChecked():
            L = self.get_L(self.L_edit)
            signal = pyboat.normalize_with_envelope(signal, L, dt=self.dt)

        self.w_position += 20

        self.anaWindows[self.w_position] = WaveletAnalyzer(
            signal=signal,
            dt=self.dt,
            Tmin=wlet_pars["T_min"],
            Tmax=wlet_pars["T_max"],
            pow_max=wlet_pars["p_max"],
            step_num=wlet_pars["step_num"],
            position=self.w_position,
            signal_id="Synthetic Signal",
            time_unit=self.time_unit,
            DEBUG=self.debug,
            parent=self,
        )

    def run_fourier_ana(self):
        if not np.any(self.raw_signal):
            self.NoSignalSelected = MessageWindow(
                "Please select a signal first!", "No Signal"
            )
            return False

        # shift new analyser windows
        self.w_position += 20

        if self.cb_use_detrended2.isChecked():
            trend = self.calc_trend()
            if trend is not None:
                signal = self.raw_signal - trend
            else:
                return
        else:
            signal = self.raw_signal

        if self.cb_use_envelope2.isChecked():
            L = self.get_L(self.L_edit)
            if L is not None:
                signal = pyboat.normalize_with_envelope(signal, L, self.dt)
            else:
                return

        # periods or frequencies?
        if self.cb_FourierT.isChecked():
            show_T = False
        else:
            show_T = True

        self.anaWindows[self.w_position] = FourierAnalyzer(
            signal=signal,
            dt=self.dt,
            signal_id="Synthetic Signal",
            position=self.w_position,
            time_unit=self.time_unit,
            show_T=show_T,
            parent=self,
        )
